# Copyright 2022 Inductor, Inc.

"""Abstractions over table-structured storage systems.

For example, table-structured storage systems include relational databases, data
warehouses, data lakes, data lakehouses, and document stores.
"""

import abc
import string
import sys
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd


# Type for rows in a Table.
Row = Dict[str, Any]


def _interpolate_sql(
    raw_str: str,
    placeholder: str) -> Tuple[str, Tuple[Any]]:
    """Returns SQL string interpolated.

    If raw_str contains segments enclosed in curly braces, interpolates the
    given string using the values of the variables in the stack frame, three
    frames above (via string.Formatter).

    Args:
        raw_str: Raw string to interpolate.
        placeholder: Placeholder to fill the curly braces from raw string.

    Returns:
        A tuple containing (raw_str with curly-braced segments replaced by
        placeholder, Python values of expressions contained in curly-braced
        segments of raw_str evaluated using variable values in stack frame three
        frames above).
    """
    conversions = {"a": ascii, "r": repr, "s": str}
    previous_frame = sys._getframe(3) # pylint: disable=protected-access
    result = []
    values = []
    parts = string.Formatter().parse(raw_str)
    for part in parts:
        literal_text, field_name, format_spec, conversion = part
        if literal_text:
            result.append(literal_text)
        if not field_name:
            continue
        value = eval(  # pylint: disable=eval-used
            field_name,
            previous_frame.f_globals,
            previous_frame.f_locals)
        if conversion:
            value = conversions[conversion](value)
        if format_spec:
            value = format(value, format_spec)
        if isinstance(value, (list, tuple, set)):
            result.append(",".join([placeholder] * len(value)))
            for v in value:
                values.append(v)
        else:
            result.append(placeholder)
            values.append(value)
    return ("".join(result), tuple(values))


class SqlSelectQuery:
    """Container for specification of a SQL SELECT query over a table."""

    def __init__(
        self,
        expression: str,
        after_from: str = "",
        placeholder: str = ""):
        """Initializes a SqlSelectQuery instance.

        If placeholder is not the empty string, then interpolates expression or
        after_from if either contains segments enclosed in curly braces; the
        interpolation is done using the values of the variables in the stack
        frame two frames above. The semantics of the interpolation are those of
        Python's built-in `str.format()`.

        Args:
            expression: SQL clause that will appear between SELECT and FROM.
            after_from: SQL clause that will appear after FROM clause.
            placeholder: Placeholder that will replace curly-braced segments in
                `expression` and `after_from` (generally should be placeholder
                required by the data system's Python connector library).  No
                interpolation is performed if placeholder is the empty string.
        """
        if placeholder:
            self._expression, self._expression_values = _interpolate_sql(
                expression, placeholder)
            self._after_from, self._after_from_values = _interpolate_sql(
                after_from, placeholder)
        else:
            self._expression = expression
            self._expression_values = tuple()
            self._after_from = after_from
            self._after_from_values = tuple()
        self._placeholder = placeholder

    @property
    def placeholder(self) -> str:
        """Placeholder used for interpolation."""
        return self._placeholder

    def to_sql_query_string(self, from_string: str) -> Tuple[str, Tuple[Any]]:
        """Returns SQL SELECT query string and values implied by this instance.

        Args:
            from_string: String to be used in the query's FROM clause.
        """
        return (f"SELECT {self._expression} FROM {from_string} "
                f"{self._after_from}",
                self._expression_values + self._after_from_values)


class Table(abc.ABC):
    """A set of Rows, backed by structured storage.

    All tables must allow scans.  Subclasses may provide additional modes of
    access (e.g., fast range queries, general SQL SELECT queries) and/or the
    ability to write to tables (e.g., by appending rows, by associating a row
    with a primary key).
    """

    @abc.abstractmethod
    def __iter__(self) -> Iterable[Row]:
        """Returns an iterable over this table's rows."""

    @abc.abstractmethod
    def first_row(self) -> Optional[Row]:
        """Returns first row in this table, or None if table is empty."""

    @property
    @abc.abstractmethod
    def columns(self) -> Optional[Iterable[str]]:
        """Returns this table's columns' names, if efficiently retrievable.

        Returns:
            The names of this table's columns, if efficiently retrievable from
            the underlying storage system (e.g., without iterating over
            all rows in the table); otherwise, returns None.
        """

    def column_values(self) -> Dict[str, List[Any]]:
        """Returns a Dict mapping column names to lists of column values."""
        names_to_values = {}
        for row in self:
            if not names_to_values:
                names_to_values = {k: [] for k in row.keys()}
            assert len(row) == len(names_to_values)
            for k, v in row.items():
                names_to_values[k].append(v)
        return names_to_values

    def pandas_df(self) -> pd.DataFrame:
        """Returns a Pandas DataFrame containing this table's data."""
        return pd.DataFrame(self)

    def values(self) -> List[Any]:
        """Returns a list containing the single value stored in each row.

        Requires that this table contains exactly one column.

        Raises:
            RuntimeError: if this table does not contain exactly one column.
        """
        values = []
        column_name = None
        for row in self:
            if len(row) != 1:
                raise RuntimeError(
                    "This table does not contain exactly one column.")
            cur_column_name = next(iter(row.keys()))
            if column_name is None:
                column_name = cur_column_name
            elif column_name != cur_column_name:
                raise RuntimeError(
                    "Encountered multiple different column names.")
            values.append(row[column_name])
        return values

    def value(self) -> Any:
        """Returns the single value stored in this table.

        Requires that this table contains exactly one value.

        Raises:
            RuntimeError: if table does not contain exactly one row and one
                column.
        """
        v = None
        valid_data = False
        for row in self:
            if not valid_data:
                if len(row) == 1:
                    v = list(row.values())[0]
                    valid_data = True
                else:
                    break
            else:
                valid_data = False
                break
        if valid_data:
            return v
        else:
            raise RuntimeError(
                "Table does not contain exactly one row and one column.")

    @abc.abstractmethod
    def to_metadata(self) -> Dict[str, Any]:
        """Returns metadata sufficient to reconstitute this object.

        The returned value must be JSON-serializable by json.dumps().
        """

    @staticmethod
    @abc.abstractmethod
    def from_metadata(metadata: Dict[str, Any]) -> "Table":
        """Returns Table object represented by metadata.

        Args:
            metadata: Metadata representing a Table of a specific concrete
                subclass type, as produced by that subclass type's
                to_metadata() method.

        Raises:
            ValueError if metadata was not produced by the to_metadata()
            method of an instance of the class implementing this function.
        """


class SqlQueryable(Table):
    """A table that provides SQL SELECT queries."""

    @abc.abstractmethod
    def indexed_columns(self) -> Iterable[str]:
        """Returns names of columns indexed for fast queries."""

    @abc.abstractmethod
    def select(self, expression: str, after_from: str = "") -> Table:
        """Returns result of executing a SQL SELECT query on this table.

        In particular, returns the result of executing the query "SELECT
        {expression} FROM {this view} {after_from}".

        Any curly-brace-enclosed segments within expression and
        after_from are interpolated with the values resulting from
        evaluating the curly-braced segments. Lists, tuples, and sets
        are interpolated as comma-separated lists of values.

        Args:
            expression: SQL clause that will appear between SELECT and FROM.
            after_from: SQL clause that will appear after FROM clause.

        Returns:
            Table giving result set for query.
        """

    def __len__(self) -> int:
        """Returns number of rows in this table."""
        return self.select("count(*) as \"c\"").first_row()["c"]

    def shape(self) -> Tuple[int, int]:
        """Returns (number of rows, number of columns) in this table."""
        row = self.first_row()
        return (len(self), len(row) if row else 0)


class PrimaryKeyQueryable(Table):
    """A table that provides fast queries by unique primary key."""

    @abc.abstractmethod
    def primary_key_column(self) -> str:
        """Returns name of column containing primary key."""

    @abc.abstractmethod
    def get(self, key: Any, default: Optional[Row] = None) -> Optional[Row]:
        """Returns row having given primary key.

        Args:
            key: Primary key value.
            default: Value returned if no row having given primary key exists.

        Returns:
            Row having given primary key if such a row exists; otherwise,
            returns default.
        """

    def __getitem__(self, key: Any) -> Row:
        """Returns row having given primary key.

        Args:
            key: Primary key value.

        Raises:
            KeyError: if no row exists having given primary key.
        """
        row = self.get(key)
        if row is not None:
            return row
        else:
            raise KeyError("No row exists for given key.")

    @abc.abstractmethod
    def __contains__(self, key: Any) -> bool:
        """Returns True if table contains given primary key, False otherwise.

        Args:
            key: Primary key value.
        """


class Appendable(Table):
    """A table to which rows can be appended."""

    @abc.abstractmethod
    def extend(self, rows: Iterable[Row]):
        """Adds given rows to this table.

        Args:
            rows: Rows to be added to this table.
        """

    def append(self, row: Row):
        """Adds given row to this table.

        Args:
            row: Row to be added to this table.
        """
        self.extend([row])


class PrimaryKeyWritable(Appendable):
    """A table providing fast row insert, update, and delete by primary key."""

    @abc.abstractmethod
    def primary_key_column(self) -> str:
        """Returns name of column containing primary key."""

    @abc.abstractmethod
    def set(self, key: Any, row: Row, skip_if_exists: bool = False) -> bool:
        """Replaces row having given primary key with given row values.

        If no row currently exists having given primary key, then a new row is
        inserted having that primary key and given row values.  Otherwise, if
        skip_if_exists is False, the existing row is replaced with the given
        row values.

        Args:
            key: Primary key value.
            row: Row values.
            skip_if_exists: If True, then this method does nothing if a row
                having the given primary key value is already present in this
                table.

        Returns:
            True if row was inserted or replaced, and False otherwise.
        """

    def __setitem__(self, key: Any, row: Row):
        """Replaces row having given primary key with given row values.

        If no row currently exists having given primary key, then a new row is
        inserted having that primary key and given row values.  Otherwise, the
        existing row is replaced with the given row values.

        Alias for set(key, row, skip_if_exists=False).

        Args:
            key: Primary key value.
            row: Row values.
        """
        self.set(key, row, skip_if_exists=False)

    @abc.abstractmethod
    def update(self, key: Any, values: Row):
        """Updates row having given primary key to contain given values.

        If no row currently exists having given primary key, then a new row is
        inserted having that primary key and the given values.  Otherwise, the
        existing row is updated to have the given values; only columns
        explicitly referenced in values are updated.

        Args:
            key: Primary key value.
            values: Values to assign to a subset of columns.
        """

    @abc.abstractmethod
    def __delitem__(self, key: Any):
        """Deletes row having given primary key.

        Nothing occurs if no row exists having given primary key.

        Args:
            key: Primary key value.
        """

    @abc.abstractmethod
    def increment(
        self,
        key: Any,
        column_name: str,
        increment_by: int) -> Union[int, float]:
        """Increments specified column in row having given primary key.

        The increment is performed atomically.

        Args:
            key: Primary key value.
            column_name: Name of column to be incremented.
            increment_by: Value by which to increment.

        Returns:
            The new value of the column after it has been incremented.
        """


class WithPrimaryKey(PrimaryKeyQueryable, PrimaryKeyWritable):
    """Type for tables that are PrimaryKeyQueryable and PrimaryKeyWritable."""



# Copyright 2022 Inductor, Inc.

"""Definitions of environments in which Inductor apps can run."""

class Env:
    """Parent for classes representing Inductor app environments."""


class Local(Env):
    """Environment representing a local workstation."""

    def __init__(
        self,
        base_dir: str,
        *,
        http_host: str = "127.0.0.1",
        http_port: int = 8000):
        """Constructs a new Local instance.

        Args:
            base_dir: Directory on the local filesystem to be used by Inductor
                for storage.  Will be created if does not already exist.
            http_host: Host to which Inductor's web server should bind.
            http_port: Port to which Inductor's web server should bind.
        """
        self.base_dir = base_dir
        self.http_host = http_host
        self.http_port = http_port


class Server(Env):
    """Environment representing a server (e.g., container/VM) in the cloud."""

    def __init__(
        self,
        app_name: str,
        database_url: str,
        object_store_url: str,
        *,
        http_host: str = "0.0.0.0",
        http_port: int = 80):
        # pylint: disable=anomalous-backslash-in-string
        """Constructs a new instance.

        Args:
            app_name: Name for app in this environment.  Should uniquely
                identify app among all apps run using Server environment
                with the same database_url.
            database_url: URL identifying database to be used for storage of
                app metadata and Inductor-created tables.  This database should
                be persistent independently of (and accessible from) the compute
                node or container used to run the app.  The URL should have
                format
                    system://username:password@host:port/database
                where ``system`` is either mysql or postgresql.  The database
                role identified by username should have the ability to create
                and write to (as well as read) tables in the specified database.
            object_store_url: URL identifying object storage location to be
                used for storage of app metadata and data.  This storage
                location should be persisent independently of (and accessible
                from) the server (e.g., compute node or container) used to run
                the app.  The URL should have the following format\:

                - If using a filesystem mounted into node/container (e.g., via
                  Amazon Elastic File System): ``file://path/to/use/``
                - If using AWS S3: ``s3://bucket/prefix/``,
                  where prefix is optional.
                - If using Google Cloud Storage (GCS): ``gs://bucket/prefix/``,
                  where prefix is optional.
            http_host: Host to which Inductor's web server should bind.
            http_port: Port to which Inductor's web server should bind.
        """
        # pylint: enable=anomalous-backslash-in-string
        # Validate arguments
        if not database_url.startswith(("mysql", "postgresql", "postgres")):
            raise ValueError(
                "database_url must start with mysql:// or postgresql://")
        if not object_store_url.startswith(("file://", "s3://", "gs://")):
            raise ValueError(
                "object_store_url must start with file://, s3://, or gs://")
        # Set instance fields
        self.app_name = app_name
        self.database_url = database_url
        self.object_store_url = object_store_url
        self.http_host = http_host
        self.http_port = http_port


class InductorCloud(Local):
    """Environment representing Inductor's cloud deployment service."""

    def __init__(
        self,
        app_name: str):
        """Constructs a new InductorCloud instance.

        Args:
            app_name: Name for app in this environment.  Should uniquely
                identify app among all apps run using InductorCloud
                environment within a given deployment account.
        """
        super().__init__(
            base_dir="/home/app_data",
            http_host="0.0.0.0",
            http_port=80)
        self.app_name = app_name



# Copyright 2022 Inductor, Inc.

"""Abstractions for DynamoDB tables."""

import datetime
import decimal
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

import boto3
from botocore import exceptions

from inductor.data.table import table


# Supported Python types for primary key columns in this implementation.
PrimaryKeyColumnType = Union[str, int, decimal.Decimal, bytes]
# Supported types for keys passed to this module's public
# methods as parameters.
PrimaryKeyType = Union[
    PrimaryKeyColumnType,
    Tuple[PrimaryKeyColumnType, PrimaryKeyColumnType],
    List[PrimaryKeyColumnType]
]


def _standardize_input(value: Any) -> Any:
    """Converts a value to be able to be stored in DynamoDB.

    DynamoDB does not support all Python data types. This method converts
    values to a data type that is compatible with being
    inserted into DynamoDB, if possible. This method is the reverse
    of _standardize_output().

    This method adds support for inserting the following data
    types into DynamoDB:
        - float
        - datetime.date
        - datetime.datetime

    Args:
        value: The value to standardize.

    Returns:
        The standardized value.
    """
    if isinstance(value, float):
        return decimal.Decimal(str(value))
    elif isinstance(value, (datetime.date, datetime.datetime)):
        return value.isoformat()
    elif isinstance(value, dict):
        return {k: _standardize_input(v)
                for k, v in value.items()}
    elif isinstance(value, set):
        return set(_standardize_input(item) for item in value)
    elif isinstance(value, list):
        return [_standardize_input(v) for v in value]
    else:
        return value


def _standardize_output(value: Any) -> Any:
    """Converts values from DynamoDB to their original Python data types.

    DynamoDB does not support all Python data types. This method converts
    values to their original Python data types. This method is based on
    the reverse of _standardize_input().

    This method adds support for getting the following data
    types from DynamoDB:
        - float
        - datetime.date
        - datetime.datetime

    Args:
        value: The value to standardize.

    Returns:
        The standardized value.
    """
    if isinstance(value, decimal.Decimal):
        converted_value = str(value)
        if "." in converted_value:
            return float(converted_value)
        else:
            return int(converted_value)
    elif isinstance(value, str):
        try:
            return datetime.date.fromisoformat(value)
        except ValueError:
            pass
        try:
            return datetime.datetime.fromisoformat(value)
        except ValueError:
            pass
        return value
    elif isinstance(value, boto3.dynamodb.types.Binary):
        return value.value
    elif isinstance(value, dict):
        return {
            k: _standardize_output(v)
            for k, v in value.items()}
    elif isinstance(value, set):
        return set(
            _standardize_output(item)
            for item in value)
    elif isinstance(value, list):
        return [
            _standardize_output(item)
            for item in value]
    else:
        return value


class DynamoDBTable(table.WithPrimaryKey):
    """A Table backed by a DynamoDB table."""

    def __init__(
        self,
        table_name: str,
        partition_key_column: str,
        sort_key_column: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
        endpoint_url: Optional[str] = None,
    ):
        """Constructs a new DynamoDBTable instance.

        Args:
            table_name: Name of DynamoDB table.
            partition_key_column: Name of column containing partition key
                (i.e., name of DynamoDB partition key attribute). Values
                in the partition key column must be of type str, int, or bytes.
            sort_key_column: Optional name of column containing sort key
                (i.e., name of DynamoDB sort key attribute). Values in the
                sort key column must be of type str, int, or bytes.
            tags: Optional AWS resource tags to be applied when creating
                the table, if the table does not already exist.
            endpoint_url: Optional Endpoint URL to be used to communicate
                with the DynamoDB service.  Intended to be used to connect
                with a local DynamoDB instance for development and testing
                purposes.  In general, should not be provided if utilizing
                an actual AWS account.
        """
        self._table_name = table_name
        self._partition_key_column = partition_key_column
        self._sort_key_column = sort_key_column
        self._tags = tags
        self._endpoint_url = endpoint_url

        self._boto_dyn_resource = boto3.resource(
            "dynamodb", endpoint_url=self._endpoint_url
        )

        # Load the DynamoDB table.
        # If it does not exist, do not create the table until the start of the
        #     first method that adds an item to the table. This is to avoid
        #     creating a table with the incorrect key schema.
        self._load_boto_dyn_table()

    def _table_exists(self) -> bool:
        """Checks if the DynamoDB table exists.

        If the table has not been loaded, attempts to load the table.

        Returns:
            True if the table exists, False otherwise.
        """
        return self._boto_dyn_table is not None or self._load_boto_dyn_table()

    def _load_boto_dyn_table(self) -> bool:
        """Loads the DynamoDB table if it exists.

        Sets self._boto_dyn_table to the Boto3 DynamoDB Table
        Resource if it exists, otherwise sets it to None.
        If the table exists, checks that the columns match
        the primary key.

        Returns:
            True if the table exists, False otherwise.

        Raises:
            ValueError: If the table exists but the primary key does not match.
        """
        try:
            t = self._boto_dyn_resource.Table(self._table_name)
            t.load()
        except exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "ResourceNotFoundException":
                self._boto_dyn_table = None
                return False
            else:
                raise error

        self._boto_dyn_table = t

        # Check that the columns match the primary key columns.
        boto_dyn_partition_key_column = None
        boto_dyn_sort_key_column = None
        for key_schema in self._boto_dyn_table.key_schema:
            if key_schema["KeyType"] == "HASH":
                boto_dyn_partition_key_column = key_schema["AttributeName"]
            elif key_schema["KeyType"] == "RANGE":
                boto_dyn_sort_key_column = key_schema["AttributeName"]

        if self._partition_key_column != boto_dyn_partition_key_column:
            raise ValueError(
                f"Partition key column mismatch. "
                f"Expected {self._partition_key_column} "
                f"found {boto_dyn_partition_key_column}."
            )
        if self._sort_key_column != boto_dyn_sort_key_column:
            raise ValueError(
                f"Sort key column mismatch. "
                f"Expected {self._sort_key_column} "
                f"found {boto_dyn_sort_key_column}"
            )

        return True

    def _parse_primary_key_value(
        self, key: PrimaryKeyType) -> Dict[str, PrimaryKeyColumnType]:
        """Parses a primary-key-column value and checks that it is valid.

        Args:
            key: The primary-key-column attribute value to check.
                If the table has a sort key, key must be a tuple or list in
                the form (partition_key_column value, sort_key_column value).
                Otherwise key must be in the form of a single
                partition_key_column value. Each partition_key_column value
                or sort_key_column value must be of type str, int,
                decimal.Decimal, or bytes.

        Returns:
            A dictionary containing the following keys:
                partition_key_column:
                    The partition_key_column value.
                sort_key_column (Optional):
                    The sort_key_column value, if it exists.

        Raises:
            ValueError: If the key is not valid.
        """
        if self._sort_key_column is not None:
            if not isinstance(key, (tuple, list)):
                raise ValueError(
                    f"Invalid key type. Expected tuple or list, "
                    f"found {type(key)}."
                )
            if len(key) != 2:
                raise ValueError(
                    f"Invalid key length. Expected 2, found {len(key)}."
                )
            primary_key_value = {
                self._partition_key_column: key[0],
                self._sort_key_column: key[1],
            }
        else:
            primary_key_value = {self._partition_key_column: key}

        for k, v in primary_key_value.items():
            standardized_value = _standardize_input(v)
            primary_key_value[k] = standardized_value
            if not isinstance(standardized_value,
                              (str, int, decimal.Decimal, bytes, float)):
                raise ValueError(
                    f"Primary key column value must be of type str, int, "
                    f"float/decimal.Decimal, or bytes. Found {type(v)}."
                )

        return primary_key_value

    def _validate_row(
        self, primary_key_value: Dict[str, PrimaryKeyColumnType],
        row: table.Row) -> table.Row:
        """Ensures that the given row includes the given primary key.

        If the row does not include the primary key, the primary key is
        added to the row.

        Args:
            primary_key_value: A dictionary containing the primary key
                column values in the form:
                    partition_key_column: The partition_key_column value.
                    sort_key_column (Optional): The sort_key_column value,
                        if it exists.
            row: The row to validate.

        Returns:
            The validated row.

        Raises:
            ValueError: If the row includes a primary key that is
                different from the given primary key.
        """
        for key, value in primary_key_value.items():
            if key not in row:
                row = row.copy()
                row[key] = value
            elif row[key] != value:
                raise ValueError(
                    f"Primary key column {key} mismatch. "
                    f"Expected {value} found {row[key]}."
                )
        return row

    def _create_boto_dyn_table(self, row: table.Row):
        """Creates the DynamoDB table.

        Attempts to create the DynamoDB table using the primary key
        schema implied by the given row. If the table fails to
        be created because it already exists, the table is retrieved
        from DynamoDB. A Boto3 DynamoDB table resource is stored in the
        self._boto_dyn_table attribute.

        Args:
            row: The row to use to create the table. The row is NOT
                inserted into the table.
        """
        def _get_attribute_def(key_column: str, value: Any) -> Dict[str, str]:
            map_type_to_symbol = {str: "S", int: "N",
                                  decimal.Decimal: "N", bytes: "B"}
            return {
                "AttributeName": key_column,
                "AttributeType": map_type_to_symbol[type(value)],
            }

        key_schema = [
            {"AttributeName": self._partition_key_column, "KeyType": "HASH"}]
        attribute_definitions = [
            _get_attribute_def(
                self._partition_key_column, row[self._partition_key_column]
            )
        ]
        if self._sort_key_column is not None:
            key_schema.append(
                {"AttributeName": self._sort_key_column, "KeyType": "RANGE"}
            )
            attribute_definitions.append(
                _get_attribute_def(self._sort_key_column,
                                   row[self._sort_key_column])
            )

        table_config = {
            "TableName": self._table_name,
            "KeySchema": key_schema,
            "AttributeDefinitions": attribute_definitions,
            "ProvisionedThroughput": {
                "ReadCapacityUnits": 1,
                "WriteCapacityUnits": 1,
            },
        }
        if self._tags is not None:
            table_config["Tags"] = self._tags

        try:
            self._boto_dyn_table = self._boto_dyn_resource.create_table(
                **table_config)
            self._boto_dyn_table.wait_until_exists()
        except exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "ResourceInUseException":
                # The DynamoDB table was created by another process,
                # so load the DynamoDB table resource.
                if not self._load_boto_dyn_table():
                    raise error
            else:
                raise error

    def _parse_scan_page_item(self, type_value: Dict[str, Any]) -> Any:
        """Parses the value returned by DynamoDB's scan paginator.

        Converts the given value to a Python data type based
        on its provided AttributeType. Note that this method
        does not fully convert the value to its original Python
        data type. It only converts the value to the correct
        Python data type for the AttributeType. In order to
        fully convert the value to its original Python data type,
        the value must be passed to _standardize_output(),
        after first passing through this method.

        Args:
            type_value: A single key-value pair to parse in the format:
                {AttributeType: AttributeValue}
                    where AttributeValue is initially
                    not necessarily the same Python data
                    type as denoted by AttributeType.

        Returns:
            The parsed value.
        """
        if len(type_value) != 1:
            raise ValueError(
                f"Expected a single key-value pair, found {type_value}."
            )
        for attr_type, value in type_value.items():
            if attr_type == "N":
                return decimal.Decimal(value)
            elif attr_type == "NULL":
                return None
            elif attr_type in ("SS", "NS", "BS"):
                subtype = attr_type.replace("S", "")
                return set(
                    self._parse_scan_page_item({subtype: v})
                    for v in value)
            elif attr_type == "L":
                return [
                    self._parse_scan_page_item(v)
                    for v in value]
            elif attr_type == "M":
                return {
                    k: self._parse_scan_page_item(v)
                    for k, v in value.items()}
            return value

    def __iter__(self) -> Iterable[table.Row]:
        """See base class."""
        if not self._table_exists():
            return iter([])

        boto_dyn_client = boto3.client(
            "dynamodb", endpoint_url=self._endpoint_url)

        # Scan pagination returns each item's attributes in the format:
        #   {AttributeType: {AttributeName: AttributeValue}}
        #       where AttributeValue is not necessarily the same
        #       Python type as denoted by AttributeType.
        # We need to convert this to the format:
        #   {AttributeName: AttributeValue}
        #       where AttributeValue is the same Python type as
        #       denoted by AttributeType.
        paginator = boto_dyn_client.get_paginator("scan")
        paginator = paginator.paginate(TableName=self._table_name)
        for page in paginator:
            for item in page["Items"]:
                item_dict = {}
                for key, type_value in item.items():
                    item_dict[key] = self._parse_scan_page_item(type_value)
                yield _standardize_output(item_dict)

    def first_row(self) -> Optional[table.Row]:
        """See base class."""
        if not self._table_exists():
            return None

        response = self._boto_dyn_table.scan(Limit=1)
        if "Items" in response and len(response["Items"]) > 0:
            return _standardize_output(response["Items"][0])
        else:
            return None

    @property
    def columns(self) -> Optional[Iterable[str]]:
        """See base class."""
        return None

    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""
        metadata = {}
        metadata["table_name"] = self._table_name
        metadata["partition_key_column"] = self._partition_key_column
        metadata["sort_key_column"] = self._sort_key_column
        metadata["tags"] = self._tags
        metadata["endpoint_url"] = self._endpoint_url
        return metadata

    @staticmethod
    def from_metadata(metadata: Dict[str, Any]) -> "DynamoDBTable":
        """See base class."""
        return DynamoDBTable(**metadata)

    def extend(self, rows: Iterable[table.Row]):
        """See base class."""
        standardized_rows = []
        for row in rows:
            standardized_rows.append(
                _standardize_input(row))

        if not self._table_exists():
            self._create_boto_dyn_table(standardized_rows[0])

        with self._boto_dyn_table.batch_writer() as batch:
            for row in standardized_rows:
                batch.put_item(Item=row)

    def primary_key_column(self) -> Union[str, Tuple[str, str]]:
        """See base class."""
        if self._sort_key_column is not None:
            return (self._partition_key_column, self._sort_key_column)
        else:
            return self._partition_key_column

    def get(
        self,
        key: Any,
        default: Optional[table.Row] = None) -> Optional[table.Row]:
        """See base class."""
        if not self._table_exists():
            return default

        primary_key_value = self._parse_primary_key_value(key)

        response = self._boto_dyn_table.get_item(Key=primary_key_value)

        if "Item" in response:
            return _standardize_output(response["Item"])
        else:
            return default

    def __contains__(self, key: Any) -> bool:
        """See base class."""
        if not self._table_exists():
            return False

        primary_key_value = self._parse_primary_key_value(key)

        response = self._boto_dyn_table.get_item(
            Key=primary_key_value,
            ProjectionExpression=self._partition_key_column
        )

        return "Item" in response

    def set(
        self,
        key: Any,
        row: table.Row,
        skip_if_exists: bool = False) -> bool:
        """See base class."""
        primary_key_value = self._parse_primary_key_value(key)

        standardized_row = _standardize_input(row)
        standardized_row = self._validate_row(primary_key_value,
                                              standardized_row)

        if not self._table_exists():
            self._create_boto_dyn_table(standardized_row)

        put_item = {
            "Item": standardized_row,
        }

        # To prevent a new item from replacing an existing item, use a
        # conditional expression that contains the attribute_not_exists
        # function with the name of the attribute being used as the
        # partition/sort key for the table. Since every record must contain
        # that attribute, the attribute_not_exists function will only
        # succeed if no matching item exists.
        if skip_if_exists:
            condition_expression = (
                f"attribute_not_exists({self._partition_key_column})")
            if self._sort_key_column is not None:
                condition_expression += (
                    f" AND attribute_not_exists({self._sort_key_column})")
            put_item["ConditionExpression"] = condition_expression

        try:
            self._boto_dyn_table.put_item(**put_item)
            return True
        except exceptions.ClientError as error:
            if error.response["Error"][
                "Code"] == "ConditionalCheckFailedException":
                return False
            else:
                raise error

    def update(self, key: Any, values: table.Row):
        """See base class."""
        primary_key_value = self._parse_primary_key_value(key)

        standardized_row = _standardize_input(values)
        standardized_row = self._validate_row(primary_key_value,
                                              standardized_row)

        if not self._table_exists():
            self._create_boto_dyn_table(standardized_row)

        # It is required that the primary key be excluded
        #   from the row in the update_item call.
        standardized_row = standardized_row.copy()
        del standardized_row[self._partition_key_column]
        if self._sort_key_column is not None:
            del standardized_row[self._sort_key_column]

        # If the row only contains the primary key,
        #   then use the set method instead.
        if len(standardized_row) == 0:
            self.set(key=key, row=values, skip_if_exists=True)
            return

        # ExpressionAttributeNames is required to allow
        #   the use of reserved words as column names.
        update_expression = "SET "
        expression_attribute_values = {}
        expression_attribute_names = {}
        for column, value in standardized_row.items():
            update_expression += f"#{column} = :{column}, "
            expression_attribute_values[f":{column}"] = value
            expression_attribute_names[f"#{column}"] = column
        update_expression = update_expression[:-2]

        self._boto_dyn_table.update_item(
            Key=primary_key_value,
            UpdateExpression=update_expression,
            ExpressionAttributeValues=expression_attribute_values,
            ExpressionAttributeNames=expression_attribute_names
        )

    def __delitem__(self, key: Any):
        """See base class."""
        if not self._table_exists():
            return

        primary_key_value = self._parse_primary_key_value(key)

        self._boto_dyn_table.delete_item(Key=primary_key_value)

    def increment(
        self,
        key: Any,
        column_name: str,
        increment_by: int) -> Union[int, float]:
        """See base class."""
        # Ensure that key is not None.
        if key is None:
            raise ValueError("Primary key value cannot be None.")
        # Ensure that the column_name is not the primary key column.
        if column_name in (self._partition_key_column, self._sort_key_column):
            raise ValueError("Cannot increment primary key column.")

        primary_key_value = self._parse_primary_key_value(key)
        try:
            response = self._boto_dyn_table.update_item(
                Key=primary_key_value,
                UpdateExpression=(
                    f"SET #{column_name} = #{column_name} + :increment_by"),
                ExpressionAttributeValues={":increment_by": increment_by},
                ExpressionAttributeNames={f"#{column_name}": column_name},
                ReturnValues="UPDATED_NEW"
            )
        except exceptions.ClientError as error:
            if error.response["Error"]["Code"] == "ValidationException":
                raise ValueError(
                    f"Column {column_name} does not "
                    f"exist in table for given key value.") from error
            else:
                raise error

        value_updated = str(response["Attributes"][column_name])
        if "." in value_updated:
            return float(value_updated)
        else:
            return int(value_updated)



# Copyright 2022 Inductor, Inc.

"""Abstractions for SQLAlchemy tables."""

import abc
import datetime
import re
from typing import Any, Dict, Iterable, Optional, Tuple, Union

from alembic import migration, operations
import sqlalchemy

from inductor.data.table import table


# Name of empty placeholder column used to ensure that SQLAlchemy tables never
# contain no columns
_PLACEHOLDER_COLUMN = "inductor_placeholder_column"

# Suffix appended to names of columns that thus far contain only null values.
_ALL_NULL_COLUMN_SUFFIX = "__allnull__"


class SqlAlchemyView(table.Table, abc.ABC):
    """A view of a SQLAlchemy table."""

    def __init__(
        self,
        parent: "SqlAlchemyTable",
        query: table.SqlSelectQuery,
    ):
        """Constructs a new SqlAlchemyView instance.

        Args:
            parent: The underlying SqlAlchemyTable of which this instance
                is a view.
            query: The query (over parent) defining this view.
        """
        self._parent = parent
        self._query = query

    @abc.abstractmethod
    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""

    @staticmethod
    @abc.abstractmethod
    def from_metadata(metadata: Dict[str, Any]) -> table.Table:
        """See base class."""

    def _query_string(self) -> Tuple[str, Dict[str, Any]]:
        """Returns the SQL query string and values underlying this view.

        SQLAlchemy requires that query values be contained in a dictionary,
        with query placeholders consisting of keys of that dictionary preceded
        by colons.  Therefore, this method replaces all instances of
        self._query.placeholder in the query with such keys, and constructs
        a corresponding dictionary of values.

        Returns:
            Internal query and values in a form that SQLAlchemy accepts for
            query execution.
        """
        # Get query with placeholders
        query_string, values_seq = self._query.to_sql_query_string(
            self._parent._table_name)  # pylint: disable=protected-access
        placeholder = self._query.placeholder
        if not placeholder:
            assert not values_seq
            return (query_string, {})
        assert len(values_seq) == query_string.count(placeholder)
        # Reformat query for use with SQLAlchemy
        formatted_query = ""
        formatted_values = {}
        split_query_string = query_string.split(placeholder)
        assert len(split_query_string) == len(values_seq) + 1
        colon_segments = query_string.split(":")[1:]
        i = 0
        for query_segment, value in zip(split_query_string[:-1], values_seq):
            # Generate a unique key that is not already present in query_string
            i += 1
            key = f"interpval{i}"
            while any(s.startswith(key) for s in colon_segments):
                i += 1
                key = f"interpval{i}"
            # Update formatted_query and formatted_values
            formatted_query += f"{query_segment}:{key}"
            formatted_values[key] = value
        formatted_query += split_query_string[-1]
        return (formatted_query, formatted_values)

    def __iter__(self) -> Iterable[table.Row]:
        """See base class."""
        query_string, query_values = self._query_string()
        rows = []
        with self._parent._engine.connect() as connection:  # pylint: disable=protected-access
            try:
                result = connection.execute(
                    sqlalchemy.text(query_string),
                    query_values,
                )
                for row in result:
                    row_dict = row._asdict()  # pylint: disable=protected-access
                    if self._parent._auto_ddl:
                        row_dict = self._parent._clean_row_dict(row_dict)
                    rows.append(row_dict)
                connection.commit()
            except sqlalchemy.exc.DBAPIError as error:
                connection.rollback()
                if self._parent.columns:
                    raise error
                else:
                    rows = []
        return rows.__iter__()

    def first_row(self) -> Optional[table.Row]:
        """See base class."""
        query_string, query_values = self._query_string()
        with self._parent._engine.connect() as connection:  # pylint: disable=protected-access
            try:
                result = connection.execute(
                    sqlalchemy.text(
                        f"SELECT * FROM ({query_string}) AS sq LIMIT 1"
                    ),
                    query_values,
                )
                row = result.first()
                connection.commit()
            except sqlalchemy.exc.DBAPIError as error:
                connection.rollback()
                if self._parent.columns:
                    raise error
                row = None
        if row is None:
            return None
        row_dict = row._asdict()  # pylint: disable=protected-access
        if self._parent._auto_ddl:  # pylint: disable=protected-access
            return self._parent._clean_row_dict(row_dict)  # pylint: disable=protected-access
        return row_dict

    @property
    def columns(self) -> Iterable[str]:
        """See base class."""
        row = self.first_row()
        return list(row.keys()) if row is not None else []


class SqlAlchemyTable(table.SqlQueryable, table.Appendable, abc.ABC):
    """A Table utilizing SQLAlchemy."""

    def __init__(
        self,
        table_name: str,
        sqlalchemy_url: str,
        indexed_columns: Iterable[str] = tuple(),
        auto_ddl: bool = True,
        include_index_postfix: bool = True,
    ):
        """Constructs a new SqlAlchemyTable instance.

        Args:
            table_name: Name of table in database.
            sqlalchemy_url: A SQLAlchemy-compatible database URL.
                https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls
            indexed_columns: Names of columns that should be indexed for faster
                queries.
            auto_ddl: Boolean value indicating whether or not to automatically
                execute DDL commands to create the table and add and modify
                columns. If False, then no processing of retrieved rows to
                account for prior automatic DDL is performed.
            include_index_postfix: Boolean value indicating whether or not to
                automatically include the index postfix generated by
                _get_sqlalchemy_type() when creating indices.
        """
        self._table_name = table_name
        self._engine = sqlalchemy.create_engine(sqlalchemy_url)
        self._indexed_columns = list(indexed_columns)
        self._auto_ddl = auto_ddl
        self._include_index_postfix = include_index_postfix
        if auto_ddl:
            with self._engine.connect() as connection:
                connection.execute(sqlalchemy.text(
                    f"CREATE TABLE IF NOT EXISTS {self._table_name} "
                    f"({_PLACEHOLDER_COLUMN} INTEGER)"
                ))
                connection.commit()

    @abc.abstractmethod
    def _is_undefined_column_error(
        self, error: sqlalchemy.exc.DBAPIError
    ) -> bool:
        """Returns True if error is related to an undefined column.

        Args:
            error: The SQLAlchemy error.

        Returns:
            True if the error was caused by an undefined column, otherwise
            False.
        """

    @abc.abstractmethod
    def _gen_view(self, query: table.SqlSelectQuery) -> SqlAlchemyView:
        """Creates a view over a provided query.

        Args:
            query: A SqlSelectQuery.

        Returns:
            An instance of a view class which extends SqlAlchemyView.
        """

    @abc.abstractmethod
    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""

    @staticmethod
    @abc.abstractmethod
    def from_metadata(metadata: Dict[str, Any]) -> table.Table:
        """See base class."""

    def __del__(self):
        """Disposes self._engine as necessary on object destruction."""
        self._engine.dispose()

    def __iter__(self) -> Iterable[table.Row]:
        """See base class."""
        return self._gen_view(table.SqlSelectQuery("*")).__iter__()

    def first_row(self) -> Optional[table.Row]:
        """See base class."""
        return self._gen_view(table.SqlSelectQuery("*")).first_row()

    @property
    def _columns_with_nulls(self) -> Iterable[str]:
        """Returns names of all columns except for placeholder column."""
        metadata = sqlalchemy.MetaData()
        metadata.reflect(
            self._engine, only=[self._table_name], resolve_fks=False)
        sqlalchemy_table = metadata.tables[self._table_name]
        cols = sqlalchemy_table.columns
        return [c.name for c in cols if c.name != _PLACEHOLDER_COLUMN]

    @property
    def columns(self) -> Iterable[str]:
        """See base class."""
        cols = self._columns_with_nulls
        return [c for c in cols if not c.endswith(_ALL_NULL_COLUMN_SUFFIX)]

    def indexed_columns(self) -> Iterable[str]:
        """See base class."""
        return self._indexed_columns.copy()

    def select(self, expression: str, after_from: str = "") -> SqlAlchemyView:
        """See base class."""
        # Generate placeholder that does not appear in expression,
        # after_from, or self._table_name
        i = 0
        placeholder = f"[~placeholder{i}~]"
        while (
            placeholder in expression or
            placeholder in after_from or
            placeholder in self._table_name):
            i += 1
            placeholder = f"[~placeholder{i}~]"
        # Return view
        return self._gen_view(
            table.SqlSelectQuery(
                expression=expression,
                after_from=after_from,
                placeholder=placeholder,
            ),
        )

    def _consolidate_rows(self, rows: Iterable[table.Row]) -> Dict[str, Any]:
        """Consolidates a collection of rows into a single row.

        Given a collection of rows, this method returns a single row whose
        columns are the union of all columns in the provided rows. If a
        column exists in multiple rows, the value from the first non-None
        value in the collection is used. A column that has both None and
        non-None values will never be None in the returned row.

        Args:
            rows: The rows to consolidate.

        Returns:
            A dictionary containing the consolidated row.
        """
        return_dict = {}
        for row in rows:
            for col_name, value in row.items():
                # Ensure every column is present
                # If a value for a column is non-None, ensure
                # it's present
                if (col_name not in return_dict
                    or return_dict[col_name] is None):
                    return_dict[col_name] = value
        return return_dict

    def _validate_col_name(self, col_name: Optional[str]):
        """Raises a ValueError if column name is invalid.

        Args:
            col_name: The column name to validate.

        Raises:
            ValueError if the provided column name is invalid.
        """
        if col_name == _PLACEHOLDER_COLUMN:
            raise ValueError(
                "Cannot add a column having name equal to "
                f"placeholder column name ({_PLACEHOLDER_COLUMN})."
            )
        elif col_name.endswith(_ALL_NULL_COLUMN_SUFFIX):
            raise ValueError(
                "Column names cannot end with "
                f"\"{_ALL_NULL_COLUMN_SUFFIX}\" "
                f"(encountered column name {col_name})."
            )
        elif not re.match(r"\A[a-zA-Z0-9_]+\Z", col_name):
            raise ValueError(
                "Column names must match "
                r"\A[a-zA-Z0-9_]+\Z "
                f"(encountered column name {col_name})."
            )

    def _get_sqlalchemy_type(
        self,
        value: Any,
        is_primary_key: bool,
    ) -> Tuple[sqlalchemy.types.TypeEngine, Optional[str]]:
        """Returns the SQLAlchemy type for a provided value.

        Args:
            value: A column value.
            is_primary_key: True if this column is a primary key,
                otherwise False.

        Returns:
            Tuple of (SQLAlchemy type, index postfix), where "index postfix"
            refers to a string which some databases require to accompany a
            type when that type is used to create an index.
            If the index postfix is not required by a type, the return tuple
            will be (SQLAlchemy type, None).
        """
        index_postfix = None
        if isinstance(value, bool):
            sqlalchemy_type = sqlalchemy.Boolean
        elif isinstance(value, int):
            sqlalchemy_type = sqlalchemy.Integer
        elif isinstance(value, float):
            sqlalchemy_type = sqlalchemy.Double
        elif isinstance(value, str):
            if is_primary_key:
                sqlalchemy_type = sqlalchemy.String(255)
                index_postfix = "(255)"
            else:
                sqlalchemy_type = sqlalchemy.Text
                index_postfix = "(300)"
        elif isinstance(value, bytes):
            sqlalchemy_type = sqlalchemy.LargeBinary
            index_postfix = "(600)"
        elif isinstance(value, datetime.datetime):
            sqlalchemy_type = sqlalchemy.DateTime
        elif isinstance(value, datetime.date):
            sqlalchemy_type = sqlalchemy.Date
        else:
            raise TypeError(f"Unsupported value type: {type(value)}")
        return (sqlalchemy_type, index_postfix)

    def _add_table_columns(self, rows: Iterable[table.Row]):
        """Adds columns to the underlying table for any new columns in rows.

        Args:
            rows: Rows that may contain columns not present in the underlying
                table.
        """
        if not self._auto_ddl:
            return
        existing_col_names = set(self._columns_with_nulls)
        consolidated_rows = self._consolidate_rows(rows)
        with self._engine.connect() as connection:
            ctx = migration.MigrationContext.configure(connection)
            op = operations.Operations(ctx)
            for col_name, value in consolidated_rows.items():
                self._validate_col_name(col_name)
                null_col_name = col_name + _ALL_NULL_COLUMN_SUFFIX
                is_null_present = null_col_name in existing_col_names
                should_add_col = col_name not in existing_col_names
                should_alter_col = value is not None and is_null_present
                if should_add_col or should_alter_col:
                    if value is None:
                        # Create column and null column
                        for col_to_add in [col_name, null_col_name]:
                            op.add_column(  # pylint: disable=no-member
                                self._table_name,
                                sqlalchemy.Column(
                                    col_to_add,
                                    sqlalchemy.Text,
                                    quote=False,
                                ),
                            )
                            existing_col_names.add(col_to_add)
                    else:
                        is_keyed_table = isinstance(self, SqlAlchemyKeyedTable)
                        primary_key_column = (self.primary_key_column()
                            if is_keyed_table
                            else None)
                        assert col_name is not None
                        is_primary_key = col_name == primary_key_column
                        (sqlalchemy_type, index_postfix) = (
                            self._get_sqlalchemy_type(value, is_primary_key))
                        # Null-placeholder columns should be dropped
                        if is_null_present:
                            op.drop_column(  # pylint: disable=no-member
                                self._table_name, null_col_name)
                            existing_col_names.remove(null_col_name)
                        # If column exists, drop it
                        if col_name in existing_col_names:
                            op.drop_column(self._table_name, col_name)  # pylint: disable=no-member
                        op.add_column(  # pylint: disable=no-member
                            self._table_name,
                            sqlalchemy.Column(
                                col_name,
                                sqlalchemy_type,
                                nullable=not is_primary_key,
                                quote=False,
                            ),
                        )
                        existing_col_names.add(col_name)
                        # Create index if required.
                        if col_name in self._indexed_columns:
                            include_postfix = (index_postfix is not None
                                               and self._include_index_postfix)
                            index_col_name = (
                                f"{col_name}{index_postfix}"
                                if include_postfix
                                else col_name)
                            op.create_index(  # pylint: disable=no-member
                                f"{self._table_name}__{col_name}_index",
                                self._table_name,
                                [sqlalchemy.text(index_col_name)],
                            )
                        # Add primary key if required.
                        if is_primary_key:
                            connection.execute(sqlalchemy.text(
                                f"ALTER TABLE {self._table_name} "
                                f"ADD PRIMARY KEY ({col_name})"
                            ))
            connection.commit()

    def _get_clauses_from_row(
        self,
        row: table.Row,
        omit_none_values: bool = False,
    ) -> Tuple[str, str, Dict[str, Any]]:
        """Returns columns and values clause and values dict from row.

        Args:
            row: A table row.
            omit_none_values: Boolean flag indicating whether or not columns
                and values should be added to their respective clauses if the
                value is None.

        Returns:
            Tuple of (column clause, values clause, dictionary of values).
        """
        col_names = []
        values = []
        values_dict = {}
        for i, (col, value) in enumerate(row.items()):
            if value is not None or not omit_none_values:
                key = f"val{i}"
                col_names.append(col)
                values.append(f":{key}")
                values_dict[key] = value
        col_names_clause = ",".join(col_names)
        values_clause = ",".join(values)
        return (col_names_clause, values_clause, values_dict)

    def extend(self, rows: Iterable[table.Row]):
        """See base class."""
        self._add_table_columns(rows)
        with self._engine.connect() as connection:
            columns = None
            for row in rows:
                all_values_none = all(value is None for value in row.values())
                if all_values_none:
                    if columns is None:
                        columns = self.columns
                    col_names_clause = ",".join(columns)
                    default_values = ",".join(["DEFAULT" for _ in columns])
                    connection.execute(sqlalchemy.text(
                        f"INSERT INTO {self._table_name} "
                        f"({col_names_clause}) "
                        f"VALUES({default_values})"
                    ))
                else:
                    (col_names_clause, values_clause, values) = (
                        self._get_clauses_from_row(row, omit_none_values=True))
                    query = sqlalchemy.text(
                        f"INSERT INTO {self._table_name} "
                        f"({col_names_clause}) "
                        f"VALUES({values_clause})"
                    )
                    connection.execute(query, values)
            connection.commit()

    def _standardize_value(self, value: Any) -> Any:
        """Standardizes a raw Table column value.
        
        By default, this method returns the value it receives.
        Implementations which require additional value conversions should
        override this method.
        
        Args:
            value: A raw column value from a Table row.
        
        Returns:
            A standardized value.
        """
        return value

    def _clean_row_dict(self, row_dict: table.Row) -> table.Row:
        """Cleans a raw row dictionary returned by sqlalchemy.cursors._asdict().

        By default, only _PLACEHOLDER_COLUMN and any columns ending with the
        _ALL_NULL_COLUMN_SUFFIX are removed from the return dictionary.
        Implementations which require additional cleaning should override this
        method.

        Args:
            row: A raw row dictionary returned by sqlalchemy.cursors._asdict().

        Returns:
            A new dictionary containing the cleaned contents of row_dict.
        """
        clean_row_dict = {}
        for key, value in row_dict.items():
            if key != _PLACEHOLDER_COLUMN:
                converted_value = self._standardize_value(value)
                if key.endswith(_ALL_NULL_COLUMN_SUFFIX):
                    clean_row_dict[
                        key[:-len(_ALL_NULL_COLUMN_SUFFIX)]] = converted_value
                else:
                    clean_row_dict[key] = converted_value
        return clean_row_dict

class SqlAlchemyKeyedTable(SqlAlchemyTable, table.WithPrimaryKey, abc.ABC):
    """A SqlAlchemyTable having a primary key."""

    def __init__(
        self,
        table_name: str,
        sqlalchemy_url: str,
        primary_key_column: str,
        indexed_columns: Iterable[str] = tuple(),
        auto_ddl: bool = True,
        include_index_postfix: bool = True,
    ):
        """Constructs a new SqlAlchemyKeyedTable instance.

        Args:
            table_name: Name of table in database.
            sqlalchemy_url: A SQLAlchemy-compatible database URL.
                https://docs.sqlalchemy.org/en/20/core/engines.html#database-urls
            primary_key_column: Name of column containing primary key.
            indexed_columns: Names of columns that should be indexed for faster
                queries.
            auto_ddl: Boolean value indicating whether or not to automatically
                execute DDL commands to create the table and add and modify
                columns. If False, then no processing of retrieved rows to
                account for prior automatic DDL is performed.
            include_index_postfix: Boolean value indicating whether or not to
                automatically include the index postfix generated by
                _get_sqlalchemy_type() when creating indices.
        """
        super().__init__(
            table_name=table_name,
            sqlalchemy_url=sqlalchemy_url,
            indexed_columns=indexed_columns,
            auto_ddl=auto_ddl,
            include_index_postfix=include_index_postfix,
        )
        self._primary_key_column = primary_key_column

    @property
    @abc.abstractmethod
    def _insert_skip_if_exists_suffix(self) -> str:
        """Returns INSERT statement suffix that yields skip-if-exists behavior.

        In particular, appending the returned string to an INSERT statement
        should cause the statement to skip the insert if a row already exists
        having the same primary key.
        """

    @abc.abstractmethod
    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""

    @staticmethod
    @abc.abstractmethod
    def from_metadata(metadata: Dict[str, Any]) -> table.Table:
        """See base class."""

    def extend(self, rows: Iterable[table.Row]):
        """See base class.

        All rows in rows must contain a primary key in the column named
        self.primary_key_column(). If a row in rows contains a primary key
        that already exists in the table, then the existing row in the table
        will be replaced.
        """
        for row in rows:
            if (self._primary_key_column not in row
                or row[self._primary_key_column] is None):
                raise ValueError(
                    f"All rows in rows must contain a primary key in the "
                    f"column named {self._primary_key_column}."
                )
        self._add_table_columns(rows)
        with self._engine.connect() as connection:
            for row in rows:
                delete_query = sqlalchemy.text(
                    f"DELETE FROM {self._table_name} "
                    f"WHERE {self._primary_key_column}=:key"
                )
                delete_values = {"key": row[self._primary_key_column]}
                connection.execute(delete_query, delete_values)
                (col_names_clause, values_clause, insert_values) = (
                    self._get_clauses_from_row(row, omit_none_values=True))
                insert_query = sqlalchemy.text(
                    f"INSERT INTO {self._table_name} "
                    f"({col_names_clause}) "
                    f"VALUES({values_clause})"
                )
                connection.execute(insert_query, insert_values)
            connection.commit()

    def primary_key_column(self) -> str:
        """See base class."""
        return self._primary_key_column

    def get(
        self,
        key: Any,
        default: Optional[table.Row] = None,
    ) -> Optional[table.Row]:
        """See base class."""
        with self._engine.connect() as connection:
            try:
                query = sqlalchemy.text(
                    f"SELECT * FROM {self._table_name} "
                    f"WHERE {self._primary_key_column}=:key"
                )
                values = {"key": key}
                result = connection.execute(query, values)
                table_rows = result.fetchall()
                connection.commit()
            except sqlalchemy.exc.DBAPIError as error:
                connection.rollback()
                if self._is_undefined_column_error(error):
                    columns = self.columns
                    if columns or columns is None:
                        raise error
                    table_rows = []
                else:
                    raise error
        if not table_rows:
            return default
        if len(table_rows) > 1:
            raise RuntimeError("Found multiple rows having same primary key.")
        table_row = table_rows[0]
        row_dict = table_row._asdict()
        if self._auto_ddl:
            row_dict = self._clean_row_dict(row_dict)
        return row_dict

    def __contains__(self, key: Any) -> bool:
        """See base class."""
        with self._engine.connect() as connection:
            try:
                query = sqlalchemy.text(
                    f"SELECT COUNT(*) FROM {self._table_name} "
                    f"WHERE {self._primary_key_column}=:key"
                )
                values = {"key": key}
                result = connection.execute(query, values)
                return_value = result.scalar_one() > 0
                connection.commit()
                return return_value
            except sqlalchemy.exc.DBAPIError as error:
                connection.rollback()
                if self._is_undefined_column_error(error):
                    columns = self.columns
                    if columns or columns is None:
                        raise error
                    return False
                else:
                    raise error

    def set(
        self,
        key: Any,
        row: table.Row,
        skip_if_exists: bool = False,
    ) -> bool:
        """See base class.

        Primary key value cannot be None.
        """
        # Ensure that key is not None
        if key is None:
            raise ValueError("Primary key value cannot be None.")
        # Ensure that row contains primary key value
        if self._primary_key_column in row:
            if key != row[self._primary_key_column]:
                raise ValueError(
                    "Primary key value in row does not match key argument."
                )
        else:
            row = row.copy()
            row[self._primary_key_column] = key
        # Add any new table columns
        self._add_table_columns([row])
        # Insert or replace row
        with self._engine.connect() as connection:
            if skip_if_exists:
                # Check if the key exists if skip_if_exists is True
                query = sqlalchemy.text(
                    f"SELECT COUNT(*) FROM {self._table_name} "
                    f"WHERE {self._primary_key_column}=:key"
                )
                values = {"key": key}
                result = connection.execute(query, values)
                if result.scalar_one() > 0:
                    return False
            else:
                # Delete any existing row having given key if skip_if_exists
                # is False
                query = sqlalchemy.text(
                    f"DELETE FROM {self._table_name} "
                    f"WHERE {self._primary_key_column}=:key"
                )
                values = {"key": key}
                connection.execute(query, values)
            # Insert row
            (col_names_clause, values_clause, values) = (
                self._get_clauses_from_row(row, omit_none_values=True))
            query = sqlalchemy.text(
                f"INSERT INTO {self._table_name} ({col_names_clause}) "
                f"VALUES({values_clause})"
            )
            connection.execute(query, values)
            connection.commit()
        return True

    def update(self, key: Any, values: table.Row):
        """See base class.

        Primary key value cannot be None.
        """
        # Ensure that key is not None
        if key is None:
            raise ValueError("Primary key value cannot be None.")
        # Ensure that values contains a primary key value
        if self._primary_key_column in values:
            if key != values[self._primary_key_column]:
                raise ValueError(
                    "Primary key value in values does not match key argument."
                )
        else:
            values = values.copy()
            values[self._primary_key_column] = key
        # Add any new table columns
        self._add_table_columns([values])
        # Insert or update values
        with self._engine.connect() as connection:
            query = sqlalchemy.text(
                f"INSERT INTO {self._table_name} ({self._primary_key_column}) "
                f"VALUES(:key) {self._insert_skip_if_exists_suffix}"
            )
            query_values = {"key": key}
            connection.execute(query, query_values)
            if len(values) > 1:  # if values does not contain only primary key
                set_clauses = []
                query_values = {}
                for i, (col_name, col_value) in enumerate(values.items()):
                    if col_name != self._primary_key_column:
                        query_key = f"val{i}"
                        set_clauses.append(f"{col_name}=:{query_key}")
                        query_values[query_key] = col_value
                query_values["key"] = values[self._primary_key_column]
                query = sqlalchemy.text(
                    f"UPDATE {self._table_name} SET " +
                    ",".join(set_clauses) +
                    f" WHERE {self._primary_key_column}=:key")
                connection.execute(query, query_values)
            connection.commit()

    def __delitem__(self, key: Any):
        """See base class."""
        # Ensure that key is not None.
        if key is None:
            raise ValueError("Primary key value cannot be None.")
        with self._engine.connect() as connection:
            query = sqlalchemy.text(
                f"DELETE FROM {self._table_name} "
                f"WHERE {self._primary_key_column}=:key"
            )
            query_values = {"key": key}
            try:
                connection.execute(query, query_values)
                connection.commit()
            except sqlalchemy.exc.DBAPIError as error:
                connection.rollback()
                if self._is_undefined_column_error(error):
                    columns = self.columns
                    if columns or columns is None:
                        raise error
                else:
                    raise error

    def increment(
        self,
        key: Any,
        column_name: str,
        increment_by: int
    ) -> Union[int, float]:
        """See base class."""
        # Ensure that key is not None.
        if key is None:
            raise ValueError("Primary key value cannot be None.")
        # Ensure that column_name is not the primary key column.
        if column_name == self._primary_key_column:
            raise ValueError(
                "Cannot increment primary key column."
            )
        # Increment column.
        with self._engine.connect() as connection:
            try:
                update_query = sqlalchemy.text(
                    f"UPDATE {self._table_name} "
                    f"SET {column_name}={column_name}+:increment "
                    f"WHERE {self._primary_key_column}=:key"
                )
                update_values = {
                    "increment": increment_by,
                    "key": key,
                }
                connection.execute(update_query, update_values)
                select_query = sqlalchemy.text(
                    f"SELECT {column_name} FROM {self._table_name} "
                    f"WHERE {self._primary_key_column}=:key"
                )
                select_values = {"key": key}
                result = connection.execute(
                    select_query, select_values
                ).fetchone()
                if result is not None:
                    result = result._asdict()
                connection.commit()
            except sqlalchemy.exc.DBAPIError as error:
                connection.rollback()
                if self._is_undefined_column_error(error):
                    columns = self.columns
                    if columns or columns is None:
                        raise error
                    result = None
                else:
                    raise error
        if result is None or column_name not in result:
            raise ValueError(
                f"Column name {column_name} does not exist in table "
                f"{self._table_name}."
            )
        return result[column_name]



# Copyright 2022 Inductor, Inc.

"""Defines and implements a table catalog."""

import importlib
import json
import sqlite3
from typing import Any, Callable, Dict, List, Optional, Tuple

import boto3
import psycopg2
from psycopg2 import extras
import pymysql

from inductor import environment
from inductor import util
from inductor.data.table import dynamodb
from inductor.data.table import mysql
from inductor.data.table import postgresql
from inductor.data.table import sqlite
from inductor.data.table import table
from inductor.data.table import table_util


class TableCatalog:
    """A Table catalog."""

    def __init__(
        self,
        metadata_table: table.WithPrimaryKey,
        name_to_id_table: table.WithPrimaryKey,
        env: environment.Env,
        run_time: bool,
        database_url: str,
        unique_id: Callable[[], str],
        secrets: Optional[Dict[str, Any]] = None,
        physical_table_name_prefix: Optional[str] = None):
        """Constructs a new TableCatalog.

        Args:
            metadata_table: Table used to persist the metadata of the catalogued
                tables. Keyed by table ID.
            name_to_id_table: Table used to persist the mapping from table names
                to table IDs.
            env: Environment in which this catalog is to be used.
            run_time: If True, then this catalog is being used at run time.
            database_url: URL identifying database in which to store tables
                created by this catalog.  Must be in one of the formats
                accepted by table_util.table_from_url().
            unique_id: Function that returns a unique ID.
            secrets: Optionally, a dictionary of secrets to be used by tables
                created by this catalog.
            physical_table_name_prefix: Optionally, prefix to be used when
                naming tables in underlying data systems.
        """
        self._metadata_table = metadata_table
        self._name_to_id_table = name_to_id_table
        self._env = env
        self._run_time = run_time
        self._database_url = database_url
        self._unique_id = unique_id
        self._secrets = {} if secrets is None else secrets
        self._table_creation_instructions: List[
            Tuple[str, Dict[str, Any]]] = []
        self._physical_table_name_prefix = physical_table_name_prefix

    def __getitem__(self, table_name: str) -> table.Table:
        """Returns Table having given table_name in this catalog.

        Args:
            table_name: Name of a table in this catalog.

        Raises:
            KeyError if no table having given table_name exists in this
            TableCatalog.
        """
        # Get table ID from name_to_id table
        table_id = self._name_to_id_table[table_name]["id"]
        # Retrieve relevant information from metadata table
        row = self._metadata_table[table_id]
        module_qualname = row["table_module_qualname"]
        class_qualname = row["table_class_qualname"]
        metadata = json.loads(row["table_json"])
        existing_table = row["existing_table"]
        # Retrieve secrets from metadata table, secrets file, or
        # self._database_url
        if existing_table:
            secrets = self._secrets.get(table_name)
        elif class_qualname.startswith("Sqlite"):
            secrets = {}
        else:
            system, username, password, host, _, database = (
                table_util.parse_database_url(self._database_url))
            if (class_qualname.lower().startswith(system) and
                username == metadata["user"] and
                host == metadata["host"] and
                database == metadata["database"]):
                secrets = {"password": password}
        if not secrets:
            secrets = json.loads(row.get("table_secrets") or "{}")
        metadata = {**metadata, **secrets}
        # Obtain relevant table subclass
        table_subclass = importlib.import_module(module_qualname)
        for name in class_qualname.split("."):
            table_subclass = getattr(table_subclass, name)
        # Instantiate and return Table
        return table_subclass.from_metadata(metadata)

    def _up(self):
        """Executes any pre-run-time table creation or addition instructions."""
        # TODO: Remove need to "trick" the table creation functions
        # by setting self._run_time to True here.
        self._run_time = True
        for method_name, table_spec in self._table_creation_instructions:
            getattr(self, method_name)(**table_spec)
        self._run_time = False

    def table_spec(self, table_name: str) -> Dict[str, Any]:
        """Returns table_spec for table having given table_name.

        Args:
            table_name: Name of a table in this catalog.

        Raises:
            KeyError if no table having given table_name exists in this
            TableCatalog.
        """
        table_id = self._name_to_id_table[table_name]["id"]
        return json.loads(self._metadata_table[table_id]["table_spec"])

    def __contains__(self, table_name: str) -> bool:
        """Returns True if and only if table_name is present in this catalog.

        Args:
            table_name: Table name for which to determine presence.
        """
        return table_name in self._name_to_id_table

    def _add_table(
        self,
        table_name: str,
        t: table.Table,
        table_spec: Dict[str, Any],
        skip_if_exists: bool,
        existing_table: bool,
        table_secrets: Optional[Dict[str, Any]] = None) -> bool:
        """Adds references to given table t to this TableCatalog.

        If a table having given table_name already exists in this
        TableCatalog and skip_if_exists is False, then the references to it are
        replaced (the previous table is not deleted from its underlying
        storage system).

        Args:
            table_name: Name under which table is to be added to this
                TableCatalog.
            t: Table to be added to this TableCatalog.
            table_spec: Specification of this table's characteristics.
                Must be JSON-serializable.
            skip_if_exists: If True, then this method does nothing if a table
                having given table_name already exists in this catalog.
            existing_table: If True, then this table already exists in its
                underlying storage system.
            table_secrets: Optionally, secrets for this table. Must be
                JSON-serializable.

        Returns:
            True if table references were added or replaced,
                and False otherwise.
        """
        if skip_if_exists and table_name in self:
            return False

        table_id = self._unique_id()

        # Since table_id is a newly generated unique id,
        # it will never already exist in self._metadata_table.
        assert self._metadata_table.set(
            table_id,
            {
                "table_module_qualname": util.module_qualname(t.__class__),
                "table_class_qualname": t.__class__.__qualname__,
                "table_json": json.dumps(
                    t.to_metadata(), separators=(",", ":")),
                "table_spec": json.dumps(table_spec, separators=(",", ":")),
                "existing_table": existing_table,
                "table_secrets": None if table_secrets is None else
                    json.dumps(table_secrets, separators=(",", ":"))
            },
            skip_if_exists=True
        )

        return self._name_to_id_table.set(
            table_name,
            {
                "id": table_id
            },
            skip_if_exists=skip_if_exists
        )

    def create_table(
        self,
        name: str,
        *,  # all subsequent parameters are keyword-only
        faster_query_columns: Optional[List[str]] = None):
        """Creates (as necessary) a table having the given name.

        The table will be appendable and SQL-SELECT-queryable.  If a table
        having the given name already exists in this app, then it must have the
        characteristics specified by this method call's arguments, in which case
        no new table is created (i.e., in that case, this method is idempotent);
        otherwise, this method raises an exception.

        Args:
            name: Name for this table.
            faster_query_columns: Names of columns for which SQL SELECT queries
                should be accelerated (generally via creation of indexes).

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        if faster_query_columns is None:
            faster_query_columns = []
        if len(faster_query_columns) != len(set(faster_query_columns)):
            raise ValueError(
                "faster_query_columns contains duplicate column names.")
        table_spec = locals().copy()
        del table_spec["self"]
        if not self._run_time:
            self._table_creation_instructions.append(
                (self.create_table.__name__, table_spec))
        elif isinstance(self._env, (environment.Local, environment.Server)):
            if name in self:
                if self.table_spec(name) != table_spec:
                    raise RuntimeError(
                        f"Table {name} already exists but has different "
                        "table_spec.")
            else:
                if self._physical_table_name_prefix is None:
                    physical_table_name = f"{self._unique_id()}"
                else:
                    physical_table_name = (
                        f"{self._physical_table_name_prefix}"
                        f"_{self._unique_id()}")
                t = table_util.table_from_url(
                    database_url=self._database_url,
                    table_name=physical_table_name,
                    primary_key_column=None,
                    indexed_columns=faster_query_columns)
                already_exists = not self._add_table(
                    name, t, table_spec, skip_if_exists=True,
                    existing_table=False)
                if (already_exists and
                    self.table_spec(name) != table_spec):
                    raise RuntimeError(
                        f"Table {name} already exists but has different "
                        "table_spec.")
        else:
            raise TypeError(f"Unsupported env type ({type(self._env)})")

    def create_keyed_table(
        self,
        name: str,
        key_column: str,
        *,
        sql_queryable: bool = True,  # pylint: disable=unused-argument
        range_queries_only: bool = False,  # pylint: disable=unused-argument
        faster_query_columns: Optional[List[str]] = None):
        """Creates (as necessary) a keyed table having the given name.

        The table will be writable.  If a table having the given name already
        exists in this app, then it must have the characteristics specified by
        this method call's arguments, in which case no new table is created
        (i.e., in that case, this method is idempotent); otherwise, this method
        raises an exception.

        Args:
            name: Name for this table.
            key_column: Name of the column containing each row's unique key
                value.
            sql_queryable: Should the table natively support SQL SELECT queries?
            range_queries_only: If sql_queryable is True, should the table
                only natively support range queries?  Ignored if sql_queryable
                is False.
            faster_query_columns: Names of columns for which SQL SELECT queries
                should be accelerated (generally via creation of indexes).

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        if faster_query_columns is None:
            faster_query_columns = []
        if len(faster_query_columns) != len(set(faster_query_columns)):
            raise ValueError(
                "faster_query_columns contains duplicate column names.")
        table_spec = locals().copy()
        del table_spec["self"]
        if not self._run_time:
            self._table_creation_instructions.append(
                (self.create_keyed_table.__name__, table_spec))
        elif isinstance(self._env, (environment.Local, environment.Server)):
            if name in self:
                if self.table_spec(name) != table_spec:
                    raise RuntimeError(
                        f"Table {name} already exists but has different "
                        "table_spec.")
            else:
                if self._physical_table_name_prefix is None:
                    physical_table_name = f"{self._unique_id()}"
                else:
                    physical_table_name = (
                        f"{self._physical_table_name_prefix}"
                        f"_{self._unique_id()}")
                t = table_util.table_from_url(
                    database_url=self._database_url,
                    table_name=physical_table_name,
                    primary_key_column=key_column,
                    indexed_columns=faster_query_columns)
                already_exists = not self._add_table(
                    name, t, table_spec, skip_if_exists=True,
                    existing_table=False)
                if (already_exists and
                    self.table_spec(name) != table_spec):
                    raise RuntimeError(
                        f"Table {name} already exists but has different "
                        "table_spec.")
        else:
            raise TypeError(f"Unsupported env type ({type(self._env)})")

    def add_existing_sqlite_table(
        self,
        name: str,
        sqlite_file_path: str,
        sqlite_table_name: Optional[str] = None,
        primary_key_column: Optional[str] = None):
        """Adds a reference in this app to an existing sqlite table.

        If a table is already referenced in this app under the given name, then
        it must have the characteristics specified by this method call's
        arguments; otherwise, this method raises an exception.

        Args:
            name: Name in this app under which to add a reference to the
                specified existing sqlite table.
            sqlite_file_path: Path to file giving sqlite database containing
                the existing table to be added.
            sqlite_table_name: Name of existing table in file given by
                sqlite_file_path. If not provided, the value of the name
                argument is used.
            primary_key_column: If non-None, then the table will be treated as
                being keyed by the column having name given by this parameter.
                If None, then

                - If the existing table has exactly one primary key column, then
                  the table will be treated as being keyed on that column.
                - Otherwise, if the existing table has no primary key column
                  and exactly one unique index column, then the table will be
                  treated as being keyed on that column.
                - Otherwise, the table will not be treated as being keyed.

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        if sqlite_table_name is None:
            sqlite_table_name = name
        table_spec = locals().copy()
        del table_spec["self"]
        if not self._run_time:
            self._table_creation_instructions.append(
                (self.add_existing_sqlite_table.__name__, table_spec))
        elif name in self:
            if self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")
        else:
            if not primary_key_column:
                conn = sqlite3.connect(sqlite_file_path)
                conn.row_factory = sqlite3.Row
                table_columns_info = list(conn.execute(
                    f"PRAGMA table_info({sqlite_table_name})"))
                pk_columns_info = [
                    r for r in table_columns_info if r["pk"] > 0]
                if len(pk_columns_info) == 1:
                    primary_key_column = pk_columns_info[0]["name"]
                elif len(pk_columns_info) == 0:
                    table_indexes = list(conn.execute(
                        f"PRAGMA index_list({sqlite_table_name})"))
                    unique_indexes = [
                        r for r in table_indexes if r["unique"] == 1]
                    if len(unique_indexes) == 1:
                        index_name = unique_indexes[0]["name"]
                        unique_index_columns_info = list(conn.execute(
                            f"PRAGMA index_info({index_name})"))
                        if len(unique_index_columns_info) == 1:
                            primary_key_column = unique_index_columns_info[0][
                                "name"]
                conn.close()
            if primary_key_column:
                t = sqlite.SqliteKeyedTable(
                    sqlite_file_path, sqlite_table_name, primary_key_column,
                    auto_ddl=False)
            else:
                t = sqlite.SqliteTable(
                    sqlite_file_path, sqlite_table_name, auto_ddl=False)
            already_exists = not self._add_table(
                name, t, table_spec, skip_if_exists=True, existing_table=True)
            if (already_exists and
                self.table_spec(name) != table_spec):
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")

    def add_existing_mysql_table(
        self,
        name: str,
        mysql_host: str,
        mysql_user: str,
        mysql_database: str,
        mysql_table_name: Optional[str] = None,
        primary_key_column: Optional[str] = None,
        mysql_port: Optional[int] = None,
        mysql_password: Optional[str] = None):
        """Adds a reference in this app to an existing MySQL table.

        If a table is already referenced in this app under the given name, then
        it must have the characteristics specified by this method call's
        arguments; otherwise, this method raises an exception.

        Args:
            name: Name in this app under which to add a reference to the
                specified existing MySQL table.
            mysql_host: Host name of database server hosting the MySQL database
                containing the existing table to be added.
            mysql_user: MySQL database server username.
            mysql_database: Name of MySQL database containing existing table.
            mysql_table_name: Name of existing table in database given by
                preceding arguments. If not provided, the value of the name
                argument is used.
            primary_key_column: If non-None, then the table will be treated as
                being keyed by the column having name given by this parameter.
                If None, then

                - If the existing table has exactly one primary key column, then
                  the table will be treated as being keyed on that column.
                - Otherwise, if the existing table has no primary key column
                  and exactly one unique index column, then the table will be
                  treated as being keyed on that column.
                - Otherwise, the table will not be treated as being keyed.
            mysql_port: Optionally, the port on which to connect to the database
                server.
            mysql_password: Optionally, password for user.  If not provided,
                password must be provided by secrets file.

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        if mysql_table_name is None:
            mysql_table_name = name

        table_spec = locals().copy()
        del table_spec["self"]
        if not self._run_time:
            self._table_creation_instructions.append(
                (self.add_existing_mysql_table.__name__, table_spec))
            return

        del table_spec["mysql_password"]
        password = (mysql_password if mysql_password is not None
                    else self._secrets.get(name, {}).get("mysql_password"))
        if password is None:
            raise ValueError("No mysql_password was provided for "
                             f"table {name}.")

        if name in self:
            if self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")
        else:
            if not primary_key_column:
                conn_params = {
                    "host": mysql_host,
                    "user": mysql_user,
                    "password": password,
                    "database": mysql_database,
                    "cursorclass": pymysql.cursors.DictCursor
                }
                if mysql_port:
                    conn_params["port"] = mysql_port
                conn = pymysql.connect(**conn_params)
                with conn.cursor() as cursor:
                    cursor.execute(
                        f"SHOW KEYS FROM {mysql_table_name} "
                        "WHERE Key_name = 'PRIMARY'")
                    pk_columns = [
                        row["Column_name"] for row in cursor.fetchall()]
                if len(pk_columns) == 1:
                    primary_key_column = pk_columns[0]
                elif len(pk_columns) == 0:
                    with conn.cursor() as cursor:
                        cursor.execute(
                            f"SHOW INDEX FROM {mysql_table_name} "
                            "WHERE Non_unique = 0")
                        unique_index_columns = [
                            row["Column_name"] for row in cursor.fetchall()]
                    if len(unique_index_columns) == 1:
                        primary_key_column = unique_index_columns[0]
                conn.close()
            table_args = {
                "host": mysql_host,
                "user": mysql_user,
                "password": mysql_password,
                "database": mysql_database,
                "table_name": mysql_table_name,
                "port": mysql_port,
                "auto_ddl": False,
            }
            if primary_key_column:
                table_args["primary_key_column"] = primary_key_column
                t = mysql.MysqlKeyedTable(**table_args)
            else:
                t = mysql.MysqlTable(**table_args)
            secrets = None if mysql_password is None else {
                "mysql_password": mysql_password
            }
            already_exists = not self._add_table(
                name, t, table_spec, table_secrets=secrets, skip_if_exists=True,
                existing_table=True)
            if already_exists and self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")

    def add_existing_dynamodb_table(
        self,
        name: str,
        dynamodb_table_name: Optional[str] = None,
        dynamodb_endpoint_url: Optional[str] = None):
        """Adds a reference in this app to an existing DynamoDB table.

        If a table is already referenced in this app under the given name, then
        it must have the characteristics specified by this method call's
        arguments; otherwise, this method raises an exception.

        Args:
            name: Name in this app under which to add a reference to the
                specified existing DynamoDB table.
            dynamodb_table_name: Name of existing DynamoDB table. If not
                provided, the value of the name argument is used.
            dynamodb_endpoint_url: Optional Endpoint URL to be used to
                communicate with the DynamoDB service. Intended to be used
                to connect with a local DynamoDB instance for development
                and testing purposes. In general, should not be provided
                if utilizing an actual AWS account.

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        if dynamodb_table_name is None:
            dynamodb_table_name = name
        table_spec = locals().copy()
        del table_spec["self"]
        if not self._run_time:
            self._table_creation_instructions.append(
                (self.add_existing_dynamodb_table.__name__, table_spec))
        elif name in self:
            if self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")
        else:
            partition_key_column = None
            sort_key_column = None
            dynamodb_client = boto3.client(
                "dynamodb", endpoint_url=dynamodb_endpoint_url)
            table_description = dynamodb_client.describe_table(
                TableName=dynamodb_table_name)["Table"]
            key_schema = table_description["KeySchema"]
            for key in key_schema:
                if key["KeyType"] == "HASH":
                    partition_key_column = key["AttributeName"]
                elif key["KeyType"] == "RANGE":
                    sort_key_column = key["AttributeName"]
            assert partition_key_column is not None
            table_args = {
                "table_name": dynamodb_table_name,
                "partition_key_column": partition_key_column,
                "sort_key_column": sort_key_column,
                "endpoint_url": dynamodb_endpoint_url,
            }
            t = dynamodb.DynamoDBTable(**table_args)
            already_exists = not self._add_table(
                name, t, table_spec, skip_if_exists=True, existing_table=True)
            if (already_exists and
                self.table_spec(name) != table_spec):
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")

    def add_existing_postgresql_table(
        self,
        name: str,
        postgresql_host: str,
        postgresql_user: str,
        postgresql_database: str,
        postgresql_table_name: Optional[str] = None,
        primary_key_column: Optional[str] = None,
        postgresql_port: Optional[int] = None,
        postgresql_password: Optional[str] = None):
        """Adds a reference in this app to an existing PostgreSQL table.

        If a table is already referenced in this app under the given name, then
        it must have the characteristics specified by this method call's
        arguments; otherwise, this method raises an exception.

        Args:
            name: Name in this app under which to add a reference to the
                specified existing PostgreSQL table.
            postgresql_host: Host name of database server hosting the
                PostgreSQL database containing the existing table to be added.
            postgresql_user: PostgreSQL database server username.
            postgresql_database: Name of PostgreSQL database containing
                existing table.
            postgresql_table_name: Name of existing table in database given by
                preceding arguments. If not provided, the value of the name
                argument is used.
            primary_key_column: If non-None, then the table will be treated as
                being keyed by the column having name given by this parameter.
                If None, then

                - If the existing table has exactly one primary key column,
                  then the table will be treated as being keyed on that column.
                - Otherwise, if the existing table has no primary key column
                  and exactly one unique index column, then the table will be
                  treated as being keyed on that column.
                - Otherwise, the table will not be treated as being keyed.
            postgresql_port: Optionally, the port on which to connect to the
                database server.
            postgresql_password: Optionally, password for user.  If not
                provided, password must be provided by secrets file.

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        if postgresql_table_name is None:
            postgresql_table_name = name

        table_spec = locals().copy()
        del table_spec["self"]
        if not self._run_time:
            self._table_creation_instructions.append(
                (self.add_existing_postgresql_table.__name__, table_spec))
            return

        del table_spec["postgresql_password"]
        password = (postgresql_password if postgresql_password is not None
                    else self._secrets.get(name, {}).get("postgresql_password"))
        if password is None:
            raise ValueError("No postgresql_password was provided for "
                             f"table {name}.")

        if name in self:
            if self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")
        else:
            if not primary_key_column:
                conn_params = {
                    "host": postgresql_host,
                    "user": postgresql_user,
                    "password": password,
                    "database": postgresql_database,
                    "cursor_factory": extras.RealDictCursor
                }
                if postgresql_port:
                    conn_params["port"] = postgresql_port
                conn = psycopg2.connect(**conn_params)
                # Reference for retrieving primary key columns:
                # https://wiki.postgresql.org/wiki/Retrieve_primary_key_columns
                select_key_expression = f"""
                    SELECT a.attname, format_type(a.atttypid, a.atttypmod)
                        AS data_type
                    FROM pg_index i
                    JOIN pg_attribute a ON a.attrelid = i.indrelid
                        AND a.attnum = ANY(i.indkey)
                    WHERE i.indrelid = '"{postgresql_table_name}"'::regclass
                    """
                select_primary_key_expression = (
                    select_key_expression + "AND i.indisprimary;")
                select_unique_key_expression = (
                    select_key_expression + "AND i.indisunique;")
                with conn.cursor() as cursor:
                    cursor.execute(select_primary_key_expression)
                    if cursor.rowcount == 1:
                        primary_key_column = cursor.fetchone()["attname"]
                    else:
                        cursor.execute(select_unique_key_expression)
                        if cursor.rowcount == 1:
                            primary_key_column = cursor.fetchone()["attname"]
                conn.commit()
                conn.close()
            table_args = {
                "host": postgresql_host,
                "user": postgresql_user,
                "password": password,
                "database": postgresql_database,
                "table_name": postgresql_table_name,
                "port": postgresql_port,
                "auto_ddl": False,
            }
            if primary_key_column:
                table_args["primary_key_column"] = primary_key_column
                t = postgresql.PostgresqlKeyedTable(**table_args)
            else:
                t = postgresql.PostgresqlTable(**table_args)
            secrets = None if postgresql_password is None else {
                "postgresql_password": postgresql_password
            }
            already_exists = not self._add_table(
                name, t, table_spec, table_secrets=secrets, skip_if_exists=True,
                existing_table=True)
            if (already_exists and
                self.table_spec(name) != table_spec):
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")

    def add_existing_snowflake_table(
        self,
        name: str,
        snowflake_user: str,
        snowflake_account: str,
        snowflake_database: str,
        snowflake_schema: str,
        snowflake_warehouse: str,
        snowflake_table_name: Optional[str] = None,
        snowflake_password: Optional[str] = None):
        """Adds a reference in this app to an existing Snowflake table.

        If using this method, Inductor must have been installed with
        Snowflake support enabled, as follows:
        `pip install --upgrade inductor[snowflake]`

        If a table is already referenced in this app under the given name, then
        it must have the characteristics specified by this method call's
        arguments; otherwise, this method raises an exception.

        Args:
            name: Name in this app under which to add a reference to the
                specified existing Snowflake table.
            snowflake_user: Snowflake username.
            snowflake_account: Snowflake account name.
            snowflake_database: Snowflake database name.
            snowflake_schema: Snowflake schema name.
            snowflake_warehouse: Snowflake warehouse name to use with this
                table.
            snowflake_table_name: Name of existing table in database given by
                preceding arguments. If not provided, the value of the name
                argument is used.
            snowflake_password: Optionally, password for user.  If not
                provided, password must be provided by secrets file.

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        from inductor.data.table import snowflake  # pylint: disable=import-outside-toplevel

        if snowflake_table_name is None:
            snowflake_table_name = name

        table_spec = locals().copy()
        del table_spec["self"]
        del table_spec["snowflake"]

        if not self._run_time:
            self._table_creation_instructions.append(
                (self.add_existing_snowflake_table.__name__, table_spec))
            return

        del table_spec["snowflake_password"]
        password = (snowflake_password if snowflake_password is not None
                    else self._secrets.get(name, {}).get("snowflake_password"))
        if password is None:
            raise ValueError("No snowflake_password was provided for "
                             f"table {name}.")

        if name in self:
            if self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")
        else:
            table_args = {
                "user": snowflake_user,
                "password": password,
                "account": snowflake_account,
                "database": snowflake_database,
                "schema": snowflake_schema,
                "warehouse": snowflake_warehouse,
                "table_name": snowflake_table_name,
                "auto_ddl": False,
            }
            secrets = None if snowflake_password is None else {
                "snowflake_password": snowflake_password
            }
            t = snowflake.SnowflakeTable(**table_args)
            already_exists = not self._add_table(
                name, t, table_spec, table_secrets=secrets, skip_if_exists=True,
                existing_table=True)
            if (already_exists and
                self.table_spec(name) != table_spec):
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")

    def add_existing_bigquery_table(
        self,
        name: str,
        bigquery_dataset_name: str,
        bigquery_table_name: Optional[str] = None,
        gcp_project_id: Optional[str] = None):
        """Adds a reference in this app to an existing BigQuery table.

        If using this method, Inductor must have been installed with
        BigQuery support enabled, as follows:
        `pip install --upgrade inductor[bigquery]`

        If a table is already referenced in this app under the given name, then
        it must have the characteristics specified by this method call's
        arguments; otherwise, this method raises an exception.

        Args:
            name: Name in this app under which to add a reference to the
                specified existing BigQuery table.
            bigquery_dataset_name: Name of BigQuery dataset containing the
                table.
            bigquery_table_name: Name of existing BigQuery table.  If not
                provided, the value of the `name` argument is used.
            gcp_project_id: Optionally, the ID of the Google Cloud project
                containing the BigQuery table.  If None, then will be inferred
                from the environment in which this code is running (by Google's
                BigQuery Python library).

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        from inductor.data.table import bigquery  # pylint: disable=import-outside-toplevel

        if bigquery_table_name is None:
            bigquery_table_name = name

        table_spec = locals().copy()
        del table_spec["self"]
        del table_spec["bigquery"]

        if not self._run_time:
            self._table_creation_instructions.append(
                (self.add_existing_bigquery_table.__name__, table_spec))
            return

        if name in self:
            if self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")
        else:
            t = bigquery.BigQueryTable(
                dataset_name=bigquery_dataset_name,
                table_name=bigquery_table_name,
                project_id=gcp_project_id,
                auto_ddl=False)
            already_exists = not self._add_table(
                name, t, table_spec, skip_if_exists=True, existing_table=True)
            if (already_exists and
                self.table_spec(name) != table_spec):
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")

    def add_existing_clickhouse_table(
        self,
        name: str,
        clickhouse_host: str,
        clickhouse_user: str,
        clickhouse_database: str,
        clickhouse_table_name: Optional[str] = None,
        *,
        clickhouse_https: bool = True,
        clickhouse_port: Optional[int] = None,
        clickhouse_password: Optional[str] = None):
        """Adds a reference in this app to an existing ClickHouse table.

        If using this method, Inductor must have been installed with
        ClickHouse support enabled, as follows:
        `pip install --upgrade inductor[clickhouse]`

        If a table is already referenced in this app under the given name, then
        it must have the characteristics specified by this method call's
        arguments; otherwise, this method raises an exception.

        Args:
            name: Name in this app under which to add a reference to the
                specified existing ClickHouse table.
            clickhouse_host: Hostname of ClickHouse server hosting the
                database containing the table.
            clickhouse_user: ClickHouse server username.
            clickhouse_database: Name of ClickHouse database containing the
                table.
            clickhouse_table_name: Name of existing table in database given by
                preceding arguments.  If not provided, the value of the name
                argument is used.
            clickhouse_https: If True, then HTTPS will be used for communication
                with the ClickHouse server.  If False, then HTTP will be used.
            clickhouse_port: Optionally, the port on which to connect to the
                ClickHouse server.  If not provided, then ClickHouse's default
                port (which depends on the value of `clickhouse_https`) will
                be used.
            clickhouse_password: Optionally, password for user.  If not
                provided, password must be provided by secrets file.

        Raises:
            RuntimeError: if a table having given name already exists but does
                not have the characteristics specified by this method call's
                arguments.
        """
        # pylint: disable-next=import-outside-toplevel
        from inductor.data.table import clickhouse

        if clickhouse_table_name is None:
            clickhouse_table_name = name

        table_spec = locals().copy()
        del table_spec["self"]
        del table_spec["clickhouse"]

        if not self._run_time:
            self._table_creation_instructions.append(
                (self.add_existing_clickhouse_table.__name__, table_spec))
            return

        del table_spec["clickhouse_password"]
        password = (clickhouse_password if clickhouse_password is not None
                    else self._secrets.get(name, {}).get("clickhouse_password"))
        if password is None:
            raise ValueError("No clickhouse_password was provided for "
                             f"table {name}.")

        if name in self:
            if self.table_spec(name) != table_spec:
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")
        else:
            table_args = {
                "host": clickhouse_host,
                "username": clickhouse_user,
                "password": password,
                "database": clickhouse_database,
                "table_name": clickhouse_table_name,
                "https": clickhouse_https,
                "port": clickhouse_port,
                "auto_ddl": False
            }
            secrets = None if clickhouse_password is None else {
                "clickhouse_password": clickhouse_password
            }
            t = clickhouse.ClickHouseTable(**table_args)
            already_exists = not self._add_table(
                name, t, table_spec, table_secrets=secrets, skip_if_exists=True,
                existing_table=True)
            if (already_exists and
                self.table_spec(name) != table_spec):
                raise RuntimeError(
                    f"Table {name} already exists but has different "
                    "table_spec.")



# Copyright 2022 Inductor, Inc.

"""The Inductor CLI."""

import importlib
import multiprocessing
import os
import signal
import sys

import typer
import watchgod

import inductor

app = typer.Typer()


@app.callback()
def app_callback():
    """The Inductor CLI."""


class _InductorAppWatcher(watchgod.DefaultDirWatcher):
    """Watchgod watcher for files that define an Inductor App."""

    # Directory names to be ignored
    _ignored_dirnames = {
        ".devcontainer", ".vscode", ".git", "__pycache__", "node_modules"}

    def should_watch_dir(self, entry: os.DirEntry) -> bool:
        """See base class."""
        return (
            entry.name not in _InductorAppWatcher._ignored_dirnames and
            not entry.name.startswith("."))

    def should_watch_file(self, entry: os.DirEntry) -> bool:
        """See base class."""
        return (
            entry.name.endswith(".py") and
            not entry.name.startswith("."))


def _may_import_inductor_app(path: str, app_fullname: str) -> bool:
    """Returns True iff Python file at path may explicitly import app_fullname.

    Args:
        path: Path to Python file to test for potential import.
        app_fullname: Location of Inductor App for which to test for potential
            import, in the format "{module}:{object}".
    """
    app_module_qualname, _ = app_fullname.split(":")
    app_module_name = app_module_qualname.split(".")[-1]
    may_import = False
    with open(path) as f:
        for line in f.readlines():
            if "import " in line and app_module_name in line:
                may_import = True
                break
    return may_import


def _load_app(app_fullname: str) -> inductor.App:
    """Loads and returns Inductor App object given by app_fullname.

    Specifically, imports the Inductor App object given by app_fullname, as
    well as all Python modules in the current working directory or in packages
    in the current working directory's subtree that may explicitly import the
    Inductor App's module to ensure that all App-decorated functions are
    registered with the App object.

    Args:
        app_fullname: Location of Inductor App to load, in the format
            "{module}:{object}".

    Returns:
        The loaded Inductor App object.
    """
    app_module_qualname, app_object_qualname = app_fullname.split(":")
    # Add current working directory to sys.path
    orig_sys_path = list(sys.path)
    sys.path[0] = os.getcwd()
    # Obtain Inductor App object
    inductor_app = importlib.import_module(app_module_qualname)
    for name in app_object_qualname.split("."):
        inductor_app = getattr(inductor_app, name)
    # Import Python modules that may import app_fullname
    cwd = os.getcwd()
    in_cwd = True
    for dirpath, dirnames, filenames in os.walk(cwd):
        cur_dirname = os.path.dirname(dirpath)
        if ((in_cwd or "__init__.py" in filenames) and
            cur_dirname not in _InductorAppWatcher._ignored_dirnames and  # pylint: disable=protected-access
            not cur_dirname.startswith(".")):
            for filename in filenames:
                if (filename.endswith(".py") and
                    not filename.startswith(".") and
                    _may_import_inductor_app(
                        os.path.join(dirpath, filename), app_fullname)):
                    module_qualname, _ = os.path.splitext(filename)
                    package_path = os.path.relpath(dirpath, start=cwd)
                    if package_path != ".":
                        while package_path:
                            package_path, tail = os.path.split(package_path)
                            assert tail
                            module_qualname = tail + "." + module_qualname
                    if module_qualname.endswith(".__init__"):
                        module_qualname = module_qualname[
                            :-len(".__init__")]
                    importlib.import_module(module_qualname)
        else:
            del dirnames[:]
        in_cwd = False
    # Restore sys.path
    sys.path = orig_sys_path
    # Return the App object
    return inductor_app


def _app_up(app_fullname: str):
    """Loads and calls up() on the Inductor App given by app_fullname.

    Args:
        app_fullname: Location of Inductor App on which to call up(), in the
            format "{module}:{object}".
    """
    # Add current working directory to sys.path
    orig_sys_path = list(sys.path)
    sys.path[0] = os.getcwd()
    # Up the Inductor app
    _load_app(app_fullname).up()
    # Restore sys.path
    sys.path = orig_sys_path


@app.command()
def up(
    app_fullname: str = typer.Argument(..., metavar="APP", help=(
        "The Inductor app to start or update, "
        "in the format \"{module}:{object}\".  "
        "{object} defaults to \"app\" if not provided.")),
    reload: bool = typer.Option(False, help=(
        "Automatically reload the app whenever its code changes?  Ignored if "
        "not using a Local environment."))):
    """Start an Inductor app, or update an already running app."""
    if ":" not in app_fullname:
        app_fullname += ":app"
    if not reload:
        _app_up(app_fullname)
    else:
        app_module_qualname, _ = app_fullname.split(":")
        app_module_realpath = os.path.realpath(os.path.join(
            os.getcwd(), *app_module_qualname.split(".")) + ".py")
        process = multiprocessing.Process(target=_app_up, args=[app_fullname])
        process.start()
        for changes in watchgod.watch(
            os.getcwd(), watcher_cls=_InductorAppWatcher):
            reload_now = False
            for change_type, path in changes:
                if (change_type == watchgod.Change.deleted or
                    os.path.realpath(path) == app_module_realpath or
                    _may_import_inductor_app(path, app_fullname)):
                    reload_now = True
                    break
            if reload_now:
                if process.is_alive():
                    os.kill(process.pid, signal.SIGINT)
                    process.join(2)
                    if process.is_alive():
                        process.kill()
                        process.join(2)
                    assert not process.is_alive()
                process = multiprocessing.Process(
                    target=_app_up, args=[app_fullname])
                process.start()


if __name__ == "__main__":
    app()



# Copyright 2022 Inductor, Inc.

"""Abstractions for MySQL tables."""

from typing import Any, Dict, Iterable, Optional

import sqlalchemy

from inductor.data.table import table
from inductor.data.table import sqlalchemy as table_sa


class MysqlView(table_sa.SqlAlchemyView):
    """A view of a MySQL table."""

    def __init__(
        self,
        parent: "MysqlTable",
        query: table.SqlSelectQuery,
    ):
        """Constructs a new MysqlView instance.

        Args:
            parent: The underlying MysqlTable of which this instance
                is a view.
            query: The query (over parent) defining this view.
        """
        super().__init__(parent, query)

    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""
        raise NotImplementedError()

    @staticmethod
    def from_metadata(metadata: Dict[str, Any]) -> table.Table:
        """See base class."""
        raise NotImplementedError()


class MysqlTable(table_sa.SqlAlchemyTable):
    """A Table backed by a MySQL table."""

    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        database: str,
        table_name: str,
        indexed_columns: Iterable[str] = tuple(),
        port: Optional[int] = None,
        auto_ddl: bool = True,
        **kwargs,
    ):
        """Constructs a new MysqlTable instance.

        Args:
            host: Name of database server host hosting the MySQL database
                containing this table.
            user: MySQL database server username.
            password: Password for user.
            database: Name of database containing table.
            table_name: Name of table in database given by preceding arguments.
            indexed_columns: Names of columns that should be indexed for faster
                queries.
            port: Optionally, the port on which to connect to the database
                server.
            auto_ddl: Boolean value indicating whether or not to automatically
                execute DDL commands to create the table and add and modify
                columns.  If False, then no processing of retrieved rows to
                account for prior automatic DDL is performed.
            **kwargs: Additional parameters which should be passed to the
                constructor of the parent class.
        """
        sqlalchemy_url = sqlalchemy.URL.create(
            "mysql+pymysql",
            host=host,
            username=user,
            password=password,
            database=database,
            port=port,
        )
        super().__init__(
            **kwargs,
            table_name=table_name,
            sqlalchemy_url=sqlalchemy_url,
            indexed_columns=indexed_columns,
            auto_ddl=auto_ddl,
        )
        self._metadata = {
            "host": host,
            "user": user,
            "database": database,
            "table_name": table_name,
            "indexed_columns": list(indexed_columns),
            "auto_ddl": auto_ddl
        }
        if port is not None:
            self._metadata["port"] = port

    def _is_undefined_column_error(
        self, error: sqlalchemy.exc.DBAPIError
    ) -> bool:
        """See base class."""
        return "Unknown column" in error.args[0]

    def _gen_view(
        self,
        query: table.SqlSelectQuery,
    ) -> MysqlView:
        """See base class."""
        return MysqlView(self, query)

    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""
        table_metadata = self._metadata.copy()
        if "password" in table_metadata:
            del table_metadata["password"]
        return table_metadata

    @staticmethod
    def from_metadata(metadata: Dict[str, Any]) -> "MysqlTable":
        """See base class."""
        table_metadata = metadata.copy()
        if "mysql_password" in table_metadata:
            table_metadata["password"] = table_metadata["mysql_password"]
            del table_metadata["mysql_password"]
        return MysqlTable(**table_metadata)


class MysqlKeyedTable(MysqlTable, table_sa.SqlAlchemyKeyedTable):
    """A MysqlTable having a primary key."""

    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        database: str,
        table_name: str,
        primary_key_column: str,
        indexed_columns: Iterable[str] = tuple(),
        port: Optional[int] = None,
        auto_ddl: bool = True,
    ):
        """Constructs a new MysqlKeyedTable instance.

        Args:
            host: Name of database server host hosting the MySQL database
                containing this table.
            user: MySQL database server username.
            password: Password for user.
            database: Name of database containing table.
            table_name: Name of table in database given by preceding arguments.
            primary_key_column: Name of column containing primary key.
            indexed_columns: Names of columns that should be indexed for faster
                queries.
            port: Optionally, the port on which to connect to the database
                server.
            auto_ddl: Boolean value indicating whether or not to automatically
                execute DDL commands to create the table and add and modify
                columns.  If False, then no processing of retrieved rows to
                account for prior automatic DDL is performed.
        """
        super().__init__(
            host=host,
            user=user,
            password=password,
            database=database,
            table_name=table_name,
            indexed_columns=indexed_columns,
            port=port,
            auto_ddl=auto_ddl,
            primary_key_column=primary_key_column,
        )

    @property
    def _insert_skip_if_exists_suffix(self) -> str:
        """See base class."""
        return (
            "ON DUPLICATE KEY UPDATE "
            f"{self.primary_key_column()}={self.primary_key_column()}")

    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""
        table_metadata = super().to_metadata().copy()
        table_metadata["primary_key_column"] = self.primary_key_column()
        return table_metadata

    @staticmethod
    def from_metadata(metadata: Dict[str, Any]) -> "MysqlKeyedTable":
        """See base class."""
        table_metadata = metadata.copy()
        if "mysql_password" in table_metadata:
            table_metadata["password"] = table_metadata["mysql_password"]
            del table_metadata["mysql_password"]
        return MysqlKeyedTable(**table_metadata)



# Copyright 2022 Inductor, Inc.

"""Functionality for Inductor function execution in a local environment."""

import concurrent.futures
from typing import Any, Dict, List

from inductor.compute import common


class LocalFuture(common.Future):
    """The result of a local asynchronous call to an Inductor app function."""

    def __init__(self, concurrent_future: concurrent.futures.Future):
        """Constructs a new LocalFuture instance.

        Args:
            concurrent_future: The concurrent.futures.Future for the function
                call underlying this LocalFuture.
        """
        self._concurrent_future = concurrent_future

    def wait(self) -> Any:
        """See base class."""
        return self._concurrent_future.result()


def execute_qualname_async(
    module_qualname: str, fn_qualname: str,
    args: List[Any], kwargs: Dict[str, Any]) -> LocalFuture:
    """Asynchronously executes function given by module_qualname, fn_qualname.

    Args:
        module_qualname: Fully qualified name of module containing the function
            to be executed.
        fn_qualname: Qualified name, within module given by module_qualname,
            of function to be executed.
        args: Positional arguments to be passed to function.  Must be picklable.
        kwargs: Keyword arguments to be passed to function.  Must be picklable.

    Returns:
        A LocalFuture representing the result of asynchronously calling the
        specified function.
    """
    executor = concurrent.futures.ProcessPoolExecutor(max_workers=1)
    local_future = LocalFuture(executor.submit(
        common.execute_qualname, module_qualname, fn_qualname, args, kwargs))
    executor.shutdown(wait=False)
    return local_future



# Copyright 2022 Inductor, Inc.

"""Abstractions for Snowflake tables."""

import datetime
import re
import sys  # pylint: disable=unused-import
from typing import Any, Dict, Iterable, Optional, Tuple

import pandas as pd
from snowflake import connector

from inductor.data.table import table


# Name of empty placeholder column used to ensure that Snowflake tables never
# contain no columns
_PLACEHOLDER_COLUMN = "inductor_placeholder_column"

# Suffix appended to names of columns that thus far contain only null values.
_ALL_NULL_COLUMN_SUFFIX = "__allnull__"


def _clean_row_dict(row_dict: table.Row) -> table.Row:
    """Returns a copy of row_dict with modifications.

    Args:
        row: A raw row dictionary returned by a Snowflake connector.

    Returns:
        A new dictionary containing the same contents as row_dict, but with
        the following modifications:
            - The placeholder column, _PLACEHOLDER_COLUMN, is removed.
            - Columns with names ending in _ALL_NULL_COLUMN_SUFFIX are renamed
                to remove the suffix.
    """
    clean_row_dict = {}
    for key in row_dict.keys():
        if key != _PLACEHOLDER_COLUMN:
            if key.endswith(_ALL_NULL_COLUMN_SUFFIX):
                clean_row_dict[
                    key[:-len(_ALL_NULL_COLUMN_SUFFIX)]] = row_dict[key]
            else:
                clean_row_dict[key] = row_dict[key]
    return clean_row_dict


def _standardize_output(value: Any) -> Any:
    """Converts values from Snowflake to their original Python data types.

    Converts bytearray objects to bytes objects.
    """
    if isinstance(value, bytearray):
        return bytes(value)
    elif isinstance(value, dict):
        return {
            k: _standardize_output(v)
            for k, v in value.items()}
    elif isinstance(value, set):
        return set(
            _standardize_output(item)
            for item in value)
    elif isinstance(value, list):
        return [
            _standardize_output(item)
            for item in value]
    else:
        return value


def _format_metadata(metadata: Dict[str, Any]) -> Dict[str, Any]:
    """Formats metadata to have the right params for SnowflakeTable.

    Args:
        metadata: Metadata representing a SnowflakeTable.
    """
    table_metadata = metadata.copy()
    if "snowflake_password" in table_metadata:
        table_metadata["password"] = table_metadata["snowflake_password"]
        del table_metadata["snowflake_password"]
    return table_metadata


class SnowflakeView(table.Table):
    """A view of a Snowflake table."""

    def __init__(
        self,
        parent: "SnowflakeTable",
        query: table.SqlSelectQuery):
        """Constructs a new SnowflakeView instance.

        Args:
            parent: The underlying SnowflakeTable of which
                this instance is a view.
            query: The query (over parent) defining this view.
        """
        self._parent = parent
        self._query = query

    def _query_string(self) -> Tuple[str, Tuple[Any]]:
        """Returns the SQL query string and values underlying this view."""
        # pylint: disable=protected-access
        return self._query.to_sql_query_string(
            f"\"{self._parent._table_name}\"")

    def __iter__(self) -> Iterable[table.Row]:
        """See base class."""
        query_string, query_values = self._query_string()
        rows = []
        try:
            # pylint: disable-next=protected-access
            with self._parent._connection.cursor(
                connector.DictCursor) as cursor:
                cursor.execute(query_string, query_values)
                row_dict = cursor.fetchone()
                while row_dict is not None:
                    # pylint: disable-next=protected-access
                    if self._parent._auto_ddl:
                        row_dict = _clean_row_dict(row_dict)
                    row_dict = _standardize_output(row_dict)
                    rows.append(row_dict)
                    row_dict = cursor.fetchone()
            self._parent._connection.commit()
        except connector.errors.ProgrammingError as error:
            self._parent._connection.rollback()  # pylint: disable=protected-access
            with self._parent._connection.cursor() as cursor:  # pylint: disable=protected-access
                cursor.execute(
                    f"SELECT COUNT(*) FROM \"{self._parent._table_name}\"")  # pylint: disable=protected-access
                row_count = cursor.fetchone()[0]
            self._parent._connection.commit()  # pylint: disable=protected-access
            if row_count == 0:
                rows = []
            else:
                raise error
        return rows.__iter__()

    def first_row(self) -> Optional[table.Row]:
        """See base class."""
        try:
            # pylint: disable-next=protected-access
            with self._parent._connection.cursor(
                connector.DictCursor) as cursor:
                query_string, query_values = self._query_string()
                cursor.execute(
                    f"SELECT * FROM ({query_string}) AS subquery LIMIT 1",
                    query_values)
                row_dict = cursor.fetchone()
            # pylint: disable-next=protected-access
            self._parent._connection.commit()
        except connector.errors.ProgrammingError as error:
            self._parent._connection.rollback()  # pylint: disable=protected-access
            with self._parent._connection.cursor() as cursor:  # pylint: disable=protected-access
                cursor.execute(
                    f"SELECT COUNT(*) FROM \"{self._parent._table_name}\"")  # pylint: disable=protected-access
                row_count = cursor.fetchone()[0]
            self._parent._connection.commit()  # pylint: disable=protected-access
            if row_count == 0:
                return None
            raise error
        if row_dict is not None:
            # pylint: disable-next=protected-access
            if self._parent._auto_ddl:
                row_dict = _clean_row_dict(row_dict)
            row_dict = _standardize_output(row_dict)
        return row_dict

    @property
    def columns(self) -> Iterable[str]:
        """See base class."""
        row = self.first_row()
        return list(row.keys()) if row is not None else []

    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""
        return NotImplementedError()

    @staticmethod
    def from_metadata(metadata: Dict[str, Any]) -> table.Table:
        return NotImplementedError()


class SnowflakeTable(table.SqlQueryable, table.Appendable):
    """A Table backed by a Snowflake table."""

    def __init__(
        self,
        user: str,
        password: str,
        account: str,
        database: str,
        schema: str,
        warehouse: str,
        table_name: str,
        auto_ddl: bool = True):
        """Constructs a new SnowflakeTable instance.

        Args:
            user: Snowflake username.
            password: Password for user.
            account: Snowflake account name.
            database: Name of database containing table.
            schema: Name of schema containing table.
            warehouse: Name of warehouse to use for queries.
            table_name: Name of table in database given by preceding arguments.
            auto_ddl: Boolean value indicating whether or not to automatically
                execute DDL commands to create the table and add and modify
                columns.  If False, then no processing of retrieved rows to
                account for prior automatic DDL is performed.
        """

        self._connection_params = {
            "user": user,
            "password": password,
            "account": account,
            "autocommit": False
        }
        self._table_name = table_name
        self._database = database
        self._schema = schema
        self._warehouse = warehouse
        self._auto_ddl = auto_ddl

        self._connection = connector.connect(
            **self._connection_params)
        with self._connection.cursor() as cursor:
            cursor.execute(f"USE WAREHOUSE {warehouse}")
            cursor.execute(f"USE DATABASE {database}")
            cursor.execute(f"USE SCHEMA {schema}")
            if self._auto_ddl:
                cursor.execute(
                    f"CREATE TABLE IF NOT EXISTS \"{self._table_name}\" "
                    f"(\"{_PLACEHOLDER_COLUMN}\" INTEGER)")
        self._connection.commit()

    def __del__(self):
        """Closes self._connection as necessary on object destruction."""
        if ("sys" in globals() and hasattr("sys", "modules")):
            # To ensure that the Python interpreter is not currently exiting
            # (self._connection.close() here may raise an exception if called
            # while interpreter is exiting).
            self._connection.close()

    def __iter__(self) -> Iterable[table.Row]:
        """See base class."""
        return SnowflakeView(self, table.SqlSelectQuery("*")).__iter__()

    def first_row(self) -> Optional[table.Row]:
        """See base class."""
        return SnowflakeView(self, table.SqlSelectQuery("*")).first_row()

    @property
    def columns(self) -> Iterable[str]:
        """See base class."""
        row = self.first_row()
        return list(row.keys()) if row is not None else []

    def pandas_df(self) -> pd.DataFrame:
        """See base class."""
        with self._connection.cursor() as cursor:
            cursor.execute(f"SELECT * FROM \"{self._table_name}\"")
            df = cursor.fetch_pandas_all()
        self._connection.commit()
        return df

    def to_metadata(self) -> Dict[str, Any]:
        """See base class."""
        metadata = self._connection_params.copy()
        del metadata["autocommit"]
        del metadata["password"]
        metadata["table_name"] = self._table_name
        metadata["database"] = self._database
        metadata["schema"] = self._schema
        metadata["warehouse"] = self._warehouse
        metadata["auto_ddl"] = self._auto_ddl
        return metadata

    @staticmethod
    def from_metadata(metadata: Dict[str, Any]) -> "SnowflakeTable":
        """See base class."""
        return SnowflakeTable(**_format_metadata(metadata))

    def indexed_columns(self) -> Iterable[str]:
        """See base class."""
        return []

    def select(self, expression: str, after_from: str = "") -> SnowflakeView:
        """See base class."""
        return SnowflakeView(
            self,
            table.SqlSelectQuery(
                expression=expression, after_from=after_from, placeholder="%s"))

    def _add_table_columns(self, rows: Iterable[table.Row]):
        """Adds columns to underlying table for any new columns in rows.

        Args:
            rows: New rows based upon which to update table columns.
        """
        if not self._auto_ddl:
            return
        # Determine set of existing column names
        existing_col_names = set()
        with self._connection.cursor(
            connector.DictCursor) as cursor:
            cursor.execute(f"SELECT * FROM \"{self._table_name}\" LIMIT 0")
            for col in cursor.description:
                if col.name != _PLACEHOLDER_COLUMN:
                    existing_col_names.add(col.name)
        self._connection.commit()
        # Add table columns as necessary
        for row in rows:
            for col_name in row.keys():
                if (col_name not in existing_col_names
                    or (row[col_name] is not None and
                    col_name + _ALL_NULL_COLUMN_SUFFIX
                    in existing_col_names)):
                    if col_name == _PLACEHOLDER_COLUMN:
                        raise ValueError(
                            f"Cannot add a column having name equal to "
                            f"placeholder column name ({_PLACEHOLDER_COLUMN}).")
                    elif col_name.endswith(_ALL_NULL_COLUMN_SUFFIX):
                        raise ValueError(
                            f"Column names cannot end with "
                            f"\"{_ALL_NULL_COLUMN_SUFFIX}\" "
                            f"(encountered column name {col_name}).")
                    elif not re.match(r"\A[a-zA-Z0-9_]+\Z", col_name):
                        raise ValueError(
                            "Column names must match "
                            r"\A[a-zA-Z0-9_]+\Z "
                            f"(encountered column name {col_name}).")
                    value = row[col_name]
                    if value is None:
                        # Create two columns: col_name and col_name_all_null
                        col_name_all_null = col_name + _ALL_NULL_COLUMN_SUFFIX
                        with self._connection.cursor() as cursor:
                            cursor.execute(
                                f"ALTER TABLE \"{self._table_name}\" "
                                f"ADD COLUMN \"{col_name}\" "
                                f"TEXT DEFAULT NULL")
                            cursor.execute(
                                f"ALTER TABLE \"{self._table_name}\" "
                                f"ADD COLUMN \"{col_name_all_null}\" "
                                f"TEXT DEFAULT NULL")
                        existing_col_names.add(col_name)
                        existing_col_names.add(col_name_all_null)
                    else:
                        if isinstance(value, bool):
                            snowflake_type = "BOOLEAN"
                        elif isinstance(value, int):
                            snowflake_type = "INTEGER"
                        elif isinstance(value, float):
                            snowflake_type = "DOUBLE PRECISION"
                        elif isinstance(value, str):
                            snowflake_type = "TEXT"
                        elif isinstance(value, bytes):
                            snowflake_type = "BINARY"
                        elif isinstance(value, datetime.datetime):
                            snowflake_type = "TIMESTAMP_NTZ"
                        elif isinstance(value, datetime.date):
                            snowflake_type = "DATE"
                        else:
                            raise TypeError(
                                f"Unsupported value type: {type(value)}")
                        with self._connection.cursor() as cursor:
                            if col_name in existing_col_names:
                                cursor.execute(
                                    f"ALTER TABLE \"{self._table_name}\" "
                                    f"DROP COLUMN \""
                                    f"{col_name + _ALL_NULL_COLUMN_SUFFIX}"
                                    f"\"")
                                existing_col_names.remove(
                                    col_name + _ALL_NULL_COLUMN_SUFFIX)
                                cursor.execute(
                                    f"ALTER TABLE \"{self._table_name}\" "
                                    f"DROP COLUMN \"{col_name}\"")
                                cursor.execute(
                                    f"ALTER TABLE \"{self._table_name}\" "
                                    f"ADD COLUMN \"{col_name}\" "
                                    f"{snowflake_type} DEFAULT NULL")
                            else:
                                cursor.execute(
                                    f"ALTER TABLE \"{self._table_name}\" "
                                    f"ADD COLUMN \"{col_name}\" "
                                    f"{snowflake_type}")
                                existing_col_names.add(col_name)
                                if (col_name + _ALL_NULL_COLUMN_SUFFIX in
                                    existing_col_names):
                                    cursor.execute(
                                        f"ALTER TABLE \"{self._table_name}\" "
                                        f"DROP COLUMN "
                                        f"\""
                                        f"{col_name + _ALL_NULL_COLUMN_SUFFIX}"
                                        f"\"")
                                    existing_col_names.remove(
                                        col_name + _ALL_NULL_COLUMN_SUFFIX)
                        self._connection.commit()

    def extend(self, rows: Iterable[table.Row]):
        """See base class."""
        self._add_table_columns(rows)
        with self._connection.cursor() as cursor:
            for row in rows:
                all_values_none = all(value is None for value in row.values())
                if all_values_none:
                    cursor.execute(f"SHOW COLUMNS IN \"{self._table_name}\"")
                    col_list = cursor.fetchall()
                    default_values = ",".join(["DEFAULT" for _ in col_list])
                    cursor.execute(
                        f"INSERT INTO \"{self._table_name}\" "
                        f"VALUES ({default_values})")
                else:
                    col_names = [k for k, v in row.items() if v is not None]
                    col_names_clause = ",".join([f"\"{c}\"" for c in col_names])
                    values_clause = ",".join(["%s" for _ in col_names])
                    cursor.execute(
                        f"INSERT INTO \"{self._table_name}\" "
                        f"({col_names_clause}) "
                        f"VALUES({values_clause})",
                        tuple(row[c] for c in col_names))
        self._connection.commit()



# Copyright 2022 Inductor, Inc.

"""Abstractions for model storage and access (particularly for inference)."""

import abc
import collections
import datetime
import json
import shutil
import os
import posixpath
import re
import tempfile
import threading
from typing import Any, Dict, Optional, Tuple

import boto3
from google.cloud import storage
import joblib

from inductor import util
from inductor.data.table import table

class ModelStore(abc.ABC):
    """A persistent collection of models."""

    def __init__(
        self,
        metadata_table: table.WithPrimaryKey,
        unique_id_gen: util.UniqueIdGen):
        """Initializes ModelStore superclass' fields.

        Args:
            metadata_table: Table to be used to persist this ModelStore's
                metadata.
            unique_id_gen: UniqueIdGen to be used to generate unique IDs for
                models added to this ModelStore.  Should produce IDs that are
                guaranteed to be globally unique within the currently running
                Inductor app.
        """
        self._metadata_table = metadata_table
        self._unique_id_gen = unique_id_gen

    @abc.abstractmethod
    def _store_model(self, model: Any, model_id: str) -> Dict[str, Any]:
        """Stores model having given model_id.

        Args:
            model: The model to be stored.
            model_id: ID of the model; should be globally unique within
                the currently running Inductor app.

        Returns:
            A dictionary containing any metadata needed to load the model
            from storage, including at least the following keys:
            - framework_name: The name of the model's framework
            - framework_version: The version of the model's framework

            Additional key-value pairs may be present which are required
            to load the model; all values must be JSON-serializable.
        """

    def add_model(self, model_name: str, model: Any):
        """Adds given model to this ModelStore under given model_name.

        If a model is already present in this ModelStore under given model_name,
        then that model is replaced.

        Args:
            model_name: Name via which model will be accessible in this
                ModelStore (via __getitem__()).
            model: The model to be stored.
        """
        model_id = self._unique_id_gen.unique_id()
        storage_metadata = self._store_model(model, model_id)
        self._metadata_table[model_name] = {
            "model_id": model_id,
            "storage_metadata_json": json.dumps(storage_metadata),
            "storage_framework_name": storage_metadata["framework_name"],
            "storage_framework_version": storage_metadata[
                "framework_version"
            ],
            "python_type": str(type(model)),
            "add_time": datetime.datetime.now()
        }

    def __setitem__(self, model_name: str, model: Any):
        """Alias for self.add_model(model_name, model)."""
        self.add_model(model_name, model)

    @abc.abstractmethod
    def __getitem__(self, model_name: str) -> Optional[Any]:
        """Returns an object representing the model having given model_name.

        The type of the returned object depends on the type of model stored
        under given model_name.  In general, the returned object permits calling
        methods and retrieving the values of fields on the underlying model
        object, though the returned object may not be the underlying model
        object itself.

        If no model having given model_name is present in this ModelStore,
        then this method returns None.
        """

    def __contains__(self, model_name: str) -> bool:
        """Returns True iff this ModelStore contains a model having given name.

        Args:
            model_name: Model name for which to determine presence.
        """
        return model_name in self._metadata_table


class _CachedModelStore(ModelStore):
    """ModelStore that uses an in-memory LRU cache for model access.

    Requires implementing `_get_uncached_model` rather than
    `__getitem__`.
    """

    def __init__(
        self,
        metadata_table: table.WithPrimaryKey,
        unique_id_gen: util.UniqueIdGen
    ):
        """See base class."""
        super().__init__(metadata_table, unique_id_gen)
        self._cache_limit = 3
        self._cache = collections.OrderedDict()
        self._name_id_dict = {}
        # Ensure cache access and changes are thread-safe.
        self._cache_lock = threading.Lock()

    def __getitem__(self, model_name: str) -> Optional[Any]:
        """See base class."""
        model_metadata = self._metadata_table.get(model_name)
        if model_metadata is None:
            return None
        model_id = model_metadata["model_id"]
        with self._cache_lock:
            # If a model with this name but a different id is in the cache,
            # it should be invalidated.
            if model_name in self._name_id_dict:
                if self._name_id_dict[model_name] != model_id:
                    outdated_id = self._name_id_dict[model_name]
                    del self._cache[outdated_id]
            # If this model is stored in the cache, return it.
            if model_id in self._cache:
                self._cache.move_to_end(model_id, last=False)
                return self._cache[model_id]
        model = self._get_uncached_model(model_name)
        if model is not None:
            # Cache the model before returning it.
            with self._cache_lock:
                self._name_id_dict[model_name] = model_id
                while len(self._cache) >= self._cache_limit:
                    self._cache.popitem(last=True)
                self._cache[model_id] = model
                self._cache.move_to_end(model_id, last=False)
        return model

    @abc.abstractmethod
    def _get_uncached_model(self, model_name: str) -> Optional[Any]:
        """See base class definition of __getitem__.

        This function is called if the requested model is not found
        in the cache.
        """


class _CachedDiskModelStore(_CachedModelStore):
    """_CachedModelStore that writes/reads models to/from disk."""

    def _write_model_to_disk(
        self,
        model: Any,
        model_filepath: str,
    ) -> Dict[str, Any]:
        """Writes a provided model to disk.

        Args:
            model: The model to be written to disk.
            model_filepath: The filepath at which the model should be
                written.

        Returns:
            A key-value dictionary of model storage metadata.
        """
        model_module = model.__module__
        storage_metadata  = {}
        # Sklearn models are stored with JobLib.
        if model_module.startswith("sklearn."):
            try:
                import sklearn # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model.joblib")
                joblib.dump(model, model_path)
                storage_metadata["framework_name"] = "sklearn"
                storage_metadata["framework_version"] = sklearn.__version__
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Received scikit-learn model but the sklearn module "
                    "is not installed."
                ) from exc
        # PyTorch models are stored in two ways, natively and in TorchScript.
        elif model_module.startswith("torch."):
            try:
                import torch # pylint: disable=import-outside-toplevel
                # PyTorch models are saved in entirety and in TorchScript
                model_path = os.path.join(model_filepath, "model.pt")
                torch.save(model, model_path)
                model_ts_path = os.path.join(model_filepath, "model.ptc")
                model_scripted = torch.jit.script(model)
                model_scripted.save(model_ts_path)
                storage_metadata["framework_name"] = "torch"
                storage_metadata["framework_version"] = torch.__version__
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Received PyTorch model but the torch module "
                    "is not installed."
                ) from exc
        # Keras models are stored natively.
        elif model_module.startswith("keras."):
            try:
                import tensorflow # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model")
                model.save(model_path)
                storage_metadata["framework_name"] = "keras"
                storage_metadata["framework_version"] = tensorflow.__version__
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Received Keras model but the tensorflow module "
                    "is not installed."
                ) from exc
        # Prophet models are stored natively.
        elif model_module.startswith("prophet."):
            try:
                import prophet # pylint: disable=import-outside-toplevel
                from prophet import serialize # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model.json")
                with open(model_path, "w") as fout:
                    fout.write(serialize.model_to_json(model))
                storage_metadata["framework_name"] = "prophet"
                storage_metadata["framework_version"] = prophet.__version__
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Received Prophet model but the prophet module "
                    "is not installed."
                ) from exc
        # HuggingFace pipelines are stored natively.
        elif model_module.startswith("transformers.pipelines."):
            try:
                import transformers # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model")
                model.save_pretrained(model_path)
                storage_metadata["framework_name"] = "huggingface.pipeline"
                storage_metadata[
                    "framework_version"
                ] = transformers.__version__
                storage_metadata["hf_pipeline_task"] = model.task
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Received HuggingFace pipeline but the transformers "
                    "module is not installed."
                ) from exc
        # HuggingFace models are stored natively.
        elif model_module.startswith("transformers.models."):
            try:
                import transformers # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model")
                model.save_pretrained(model_path)
                storage_metadata["framework_name"] = "huggingface.model"
                storage_metadata[
                    "framework_version"
                ] = transformers.__version__
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Received HuggingFace model but the transformers "
                    "module is not installed."
                ) from exc
        # If not explicitly supported, attempt using joblib.
        else:
            joblib.dump(model, os.path.join(model_filepath, "model.joblib"))
            storage_metadata["framework_name"] = "joblib"
            storage_metadata["framework_version"] = ""
        assert "framework_name" in storage_metadata
        assert "framework_version" in storage_metadata
        return storage_metadata

    def _read_model_from_disk(
        self,
        model_filepath: str,
        framework_name: str,
        storage_metadata: Dict[str, Any]
    ) -> Optional[Any]:
        """Attempts to read a model from disk.

        Args:
            model_filepath: The filepath at which the model can be
                read, if the model exists.
            framework_name: The name of the framework with which
                the model was created.
            storage_metadata: The dictionary of storage metadata
                which is stored in the metadata table.

        Returns:
            The model read from disk or None if the model cannot be found.
        """
        if not os.path.exists(model_filepath):
            return None
        # Sklearn models are retrieved with JobLib.
        if framework_name == "sklearn":
            model_path = os.path.join(model_filepath, "model.joblib")
            return joblib.load(model_path)
        # PyTorch models are retrieved with native retrieval.
        elif framework_name == "torch":
            try:
                import torch # pylint: disable=import-outside-toplevel
                model_path =  os.path.join(model_filepath, "model.ptc")
                model = torch.jit.load(model_path)
                model.eval()
                return model
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Attempted to access PyTorch model but the torch module "
                    "is not installed."
                ) from exc
        # Keras models are retrieved with native retrieval.
        elif framework_name == "keras":
            try:
                import tensorflow # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model")
                return tensorflow.keras.models.load_model(model_path)
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Attempted to access Keras model but the tensorflow module "
                    "is not installed."
                ) from exc
        # Prophet models are retrieved with native retrieval.
        elif framework_name == "prophet":
            try:
                from prophet import serialize # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model.json")
                with open(model_path, "r") as fin:
                    return serialize.model_from_json(fin.read())
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Attempted to access Prophet model but the prophet module "
                    "is not installed."
                ) from exc
        # HuggingFace pipelines are retrieved natively.
        elif framework_name == "huggingface.pipeline":
            try:
                from transformers import pipeline # pylint: disable=import-outside-toplevel
                task = storage_metadata["hf_pipeline_task"]
                model_path = os.path.join(model_filepath, "model")
                return pipeline(
                    task=task, model=model_path, tokenizer=model_path,
                )
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Attempted to access HuggingFace pipeline but the "
                    "transformers module is not installed."
                ) from exc
        # HuggingFace models are retrieved natively.
        elif framework_name == "huggingface.model":
            try:
                from transformers import AutoModel # pylint: disable=import-outside-toplevel
                model_path = os.path.join(model_filepath, "model")
                return AutoModel.from_pretrained(model_path)
            except ImportError as exc:
                raise ModuleNotFoundError(
                    "Attempted to access HuggingFace model but the "
                    "transformers module is not installed."
                ) from exc
        # All unknown frameworks are retrieved with JobLib.
        else:
            model_path = os.path.join(model_filepath, "model.joblib")
            return joblib.load(model_path)


class LocalModelStore(_CachedDiskModelStore):
    """ModelStore backed by the local filesystem."""

    def __init__(
        self,
        base_path: str,
        metadata_table: table.WithPrimaryKey,
        unique_id_gen: util.UniqueIdGen
    ):
        """Constructs a new LocalModelStore.

        Args:
            base_path: Path to the directory in the local filesystem to be used
                for model storage. The directory need not already exist.
            metadata_table: Table to be used to persist this ModelStore's
                metadata.
            unique_id_gen: UniqueIdGen to be used to generate unique IDs for
                models added to this ModelStore.  Should produce IDs that are
                guaranteed to be globally unique within the currently running
                Inductor app.
        """
        super().__init__(metadata_table, unique_id_gen)
        self._base_path = base_path

    def _store_model(self, model: Any, model_id: str) -> Dict[str, Any]:
        """See base class.

        For safety (i.e., to prevent escaping base_dir), this method does not
        allow model_id to contain "..".
        """
        if ".." in model_id:
            raise ValueError("model_id cannot contain \"..\".")
        model_dir =  os.path.join(self._base_path, model_id)
        os.makedirs(model_dir)
        return self._write_model_to_disk(model, model_dir)

    def _get_uncached_model(self, model_name: str) -> Optional[Any]:
        """See base class."""
        model_metadata = self._metadata_table.get(model_name)
        if model_metadata is None:
            return None
        storage_metadata_json = model_metadata[
            "storage_metadata_json"
        ]
        storage_metadata = json.loads(storage_metadata_json)
        framework_name = model_metadata["storage_framework_name"]
        assert framework_name is not None
        model_id = model_metadata["model_id"]
        model_filepath = os.path.join(self._base_path, model_id)
        return self._read_model_from_disk(
            model_filepath,
            framework_name,
            storage_metadata
        )


class S3ModelStore(_CachedDiskModelStore):
    """ModelStore backed by AWS S3 bucket storage."""

    def __init__(
        self,
        s3_url: str,
        metadata_table: table.WithPrimaryKey,
        unique_id_gen: util.UniqueIdGen,
    ):
        """Constructs a new S3ModelStore.

        Args:
            s3_url: The S3 URL to be used for model storage, in the format
                s3://bucket/prefix/ where prefix is optional.
            metadata_table: Table to be used to persist this ModelStore's
                metadata.
            unique_id_gen: UniqueIdGen to be used to generate unique IDs for
                models added to this ModelStore.  Should produce IDs that are
                guaranteed to be globally unique within the currently running
                Inductor app.
        """
        super().__init__(metadata_table, unique_id_gen)
        (bucket_name, model_prefix) = self._get_bucket_prefix_from_url(s3_url)
        self._bucket_name = bucket_name
        self._model_prefix = model_prefix
        self._client = boto3.client("s3")

    def _get_bucket_prefix_from_url(
        self,
        s3_url: str
    ) -> Tuple[str, Optional[str]]:
        """Destructures an S3 URL into a bucket and optional prefix.

        Args:
            s3_url: An S3 URL in the format s3://bucket/prefix/
                where prefix is optional.

        Returns:
            A tuple containing the S3 bucket name and, if it exists, the
            optional prefix.
        """
        reg_exp = "\\As3://([^/]+)(/?)(.*)\\Z"
        match = re.search(reg_exp, s3_url)
        if match is None:
            raise RuntimeError("Unable to parse provided S3 URL.")
        (bucket, _, prefix) = match.groups()
        if prefix == "":
            return (bucket, None)
        else:
            return (bucket, prefix)

    def _store_model(self, model: Any, model_id: str) -> Dict[str, Any]:
        """See base class.

        For consistency with LocalModelStore, this method does not
        allow model_id to contain "..".
        """
        if ".." in model_id:
            raise ValueError("model_id cannot contain \"..\".")
        with tempfile.TemporaryDirectory() as tempdir:
            model_dir = os.path.join(tempdir, model_id)
            os.makedirs(model_dir)
            storage_metadata = self._write_model_to_disk(model, model_dir)
            # Zip the temporary directory.
            local_model_zip_name = f"{model_dir}.zip"
            model_zip_name = f"{model_id}.zip"
            if self._model_prefix is not None:
                model_zip_name = posixpath.join(
                    self._model_prefix,
                    model_zip_name
                )
            shutil.make_archive(model_dir, "zip", model_dir)
            # Upload the zipped model directory.
            with open(local_model_zip_name, "rb") as model_binary:
                self._client.upload_fileobj(
                    model_binary,
                    self._bucket_name,
                    model_zip_name
                )
            return storage_metadata

    def _get_uncached_model(self, model_name: str) -> Optional[Any]:
        """See base class."""
        model_metadata = self._metadata_table.get(model_name)
        if model_metadata is None:
            return None
        storage_metadata_json = model_metadata[
            "storage_metadata_json"
        ]
        storage_metadata = json.loads(storage_metadata_json)
        framework_name = model_metadata["storage_framework_name"]
        assert framework_name is not None
        model_id = model_metadata["model_id"]
        with tempfile.TemporaryDirectory() as tempdir:
            model_filepath = os.path.join(tempdir, model_id)
            # Download the zipped model.
            local_model_zip_name = f"{model_filepath}.zip"
            model_zip_name = f"{model_id}.zip"
            if self._model_prefix is not None:
                model_zip_name = posixpath.join(
                    self._model_prefix,
                    model_zip_name
                )
            with open(local_model_zip_name, "wb") as model_binary:
                self._client.download_fileobj(
                    self._bucket_name,
                    model_zip_name,
                    model_binary
                )
            shutil.unpack_archive(local_model_zip_name, model_filepath)
            # Read the unzipped model.
            return self._read_model_from_disk(
                model_filepath,
                framework_name,
                storage_metadata
            )


class GCSModelStore(_CachedDiskModelStore):
    """ModelStore backed by Google Cloud Storage (GCS)."""

    def __init__(
        self,
        gcs_url: str,
        metadata_table: table.WithPrimaryKey,
        unique_id_gen: util.UniqueIdGen,
    ):
        """Constructs a new GCSModelStore.

        Args:
            gcs_url: The GCS URL to be used for model storage, in the format
                gs://bucket/prefix/ where prefix is optional (and will be
                prepended to all object names, if provided).
            metadata_table: Table to be used to persist this ModelStore's
                metadata.
            unique_id_gen: UniqueIdGen to be used to generate unique IDs for
                models added to this ModelStore.  Should produce IDs that are
                guaranteed to be globally unique within the currently running
                Inductor app.
        """
        super().__init__(metadata_table, unique_id_gen)
        self._client = storage.Client()
        # Deconstruct gcs_url
        gcs_url_blob = storage.Blob.from_string(gcs_url, self._client)
        self._bucket: storage.Bucket = gcs_url_blob.bucket
        self._model_prefix = None
        if gcs_url_blob.name != "":
            self._model_prefix = gcs_url_blob.name

    def _store_model(self, model: Any, model_id: str) -> Dict[str, Any]:
        """See base class.

        For consistency with LocalModelStore, this method does not
        allow model_id to contain "..".
        """
        if ".." in model_id:
            raise ValueError("model_id cannot contain \"..\".")
        with tempfile.TemporaryDirectory() as tempdir:
            model_dir = os.path.join(tempdir, model_id)
            os.makedirs(model_dir)
            storage_metadata = self._write_model_to_disk(model, model_dir)
            # Zip the temporary directory.
            local_model_zip_name = f"{model_dir}.zip"
            model_zip_name = f"{model_id}.zip"
            if self._model_prefix is not None:
                model_zip_name = posixpath.join(
                    self._model_prefix,
                    model_zip_name
                )
            shutil.make_archive(model_dir, "zip", model_dir)
            # Upload the zipped model directory.
            blob = self._bucket.blob(model_zip_name)
            blob.upload_from_filename(
                local_model_zip_name,
                if_generation_match=0)
            return storage_metadata

    def _get_uncached_model(self, model_name: str) -> Optional[Any]:
        """See base class."""
        model_metadata = self._metadata_table.get(model_name)
        if model_metadata is None:
            return None
        storage_metadata_json = model_metadata[
            "storage_metadata_json"
        ]
        storage_metadata = json.loads(storage_metadata_json)
        framework_name = model_metadata["storage_framework_name"]
        assert framework_name is not None
        model_id = model_metadata["model_id"]
        with tempfile.TemporaryDirectory() as tempdir:
            model_filepath = os.path.join(tempdir, model_id)
            os.makedirs(model_filepath)
            # Download the zipped model.
            local_model_zip_name = f"{model_filepath}.zip"
            model_zip_name = f"{model_id}.zip"
            if self._model_prefix is not None:
                model_zip_name = posixpath.join(
                    self._model_prefix,
                    model_zip_name
                )
            blob = self._bucket.blob(model_zip_name)
            blob.download_to_filename(local_model_zip_name)
            shutil.unpack_archive(local_model_zip_name, model_filepath)
            # Read the unzipped model.
            return self._read_model_from_disk(
                model_filepath,
                framework_name,
                storage_metadata
            )
